from datetime import datetime
from json import JSONDecodeError
from typing import Dict
from typing import Mapping
from typing import Optional
from typing import Union
from urllib.parse import urljoin

import numpy as np
import pandas as pd
import requests
from google.protobuf.json_format import MessageToDict
from google.protobuf.json_format import MessageToJson
from google.protobuf.json_format import Parse
from google.protobuf.struct_pb2 import Value
from requests.exceptions import HTTPError

import tecton
from tecton import conf
from tecton._internals import errors
from tecton.framework.data_frame import FeatureVector
from tecton_core import data_types
from tecton_proto.api.featureservice.feature_service_pb2 import FeatureServerComplexDataType
from tecton_proto.api.featureservice.feature_service_pb2 import FeatureServerDataType
from tecton_proto.api.featureservice.feature_service_pb2 import GetFeaturesResponse
from tecton_proto.api.featureservice.feature_service_pb2 import GetFeaturesResult
from tecton_proto.api.featureservice.feature_service_pb2 import Metadata
from tecton_proto.api.featureservice.feature_service_pb2 import QueryFeaturesRequest
from tecton_proto.api.featureservice.feature_service_pb2 import QueryFeaturesResponse
from tecton_proto.api.featureservice.feature_service_request_pb2 import GetFeaturesRequest
from tecton_spark.request_context import RequestContext


TYPE_BOOLEAN = "boolean"
TYPE_FLOAT64 = "float64"
TYPE_INT64 = "int64"
TYPE_STRING = "string"
TYPE_STRING_ARRAY = "string_array"
TYPE_NULL_VALUE = "null_value"
TYPE_ERROR = "error"


class _QueryHelper:
    def __init__(
        self,
        workspace_name: str,
        feature_service_name: Optional[str] = None,
        feature_view_name: Optional[str] = None,
    ):
        assert (feature_service_name is not None) ^ (feature_view_name is not None)
        self.workspace_name = workspace_name
        self.feature_service_name = feature_service_name
        self.feature_view_name = feature_view_name

    def _prepare_headers(self) -> Dict[str, str]:
        token = conf.get_or_none("TECTON_API_KEY")
        if not token:
            raise errors.FS_API_KEY_MISSING

        return {"authorization": f"Tecton-key {token}"}

    def query_features(self, join_keys: Mapping[str, Union[int, np.int_, str, bytes]]) -> "tecton.TectonDataFrame":
        """
        Queries the FeatureService with partial set of join_keys defined in the OnlineServingIndex
        of the enclosed feature definitions. Returns feature vectors for all matched records.
        See OnlineServingIndex.

        :param join_keys: Query join keys, i.e., a union of join keys in OnlineServingIndex of all
            enclosed feature definitions.
        :return: A TectonDataFrame
        """
        request = QueryFeaturesRequest()
        self._prepare_request_params(request.params, join_keys)
        http_response = requests.post(
            urljoin(conf.get_or_raise("FEATURE_SERVICE") + "/", "v1/feature-service/query-features"),
            data=MessageToJson(request),
            headers=self._prepare_headers(),
        )

        self._detailed_http_raise_for_status(http_response)

        response = QueryFeaturesResponse()
        Parse(http_response.text, response, True)

        pandas_df = self._query_response_to_pandas(response, join_keys)

        import tecton

        return tecton.TectonDataFrame._create(pandas_df)

    def get_feature_vector(
        self,
        join_keys: Mapping[str, Union[int, np.int_, str, bool]],
        include_join_keys_in_response: bool,
        request_context_map: Mapping[str, Union[int, np.int_, str, float, bool]],
        request_context_schema: RequestContext,
    ) -> FeatureVector:
        """
        Returns a single Tecton FeatureVector.

        :param join_keys: Join keys of the enclosed feature definitions.
        :param include_join_keys_in_response: Whether to include join keys as part of the response FeatureVector.
        :param request_context_map: Dictionary of request context values.

        :return: A FeatureVector of the results.
        """
        request = GetFeaturesRequest()
        self._prepare_request_params(request.params, join_keys, request_context_map, request_context_schema)

        http_response = requests.post(
            urljoin(conf.get_or_raise("FEATURE_SERVICE") + "/", "v1/feature-service/get-features"),
            data=MessageToJson(request),
            headers=self._prepare_headers(),
        )

        self._detailed_http_raise_for_status(http_response)

        response = GetFeaturesResponse()
        Parse(http_response.text, response, True)

        return self._response_to_feature_vector(response, include_join_keys_in_response, join_keys)

    def _response_to_feature_vector(
        self,
        response: GetFeaturesResponse,
        include_join_keys: bool,
        join_keys: Dict,
    ) -> FeatureVector:
        features = {}
        if include_join_keys:
            for k, v in join_keys.items():
                features[k] = v

        features.update(self._feature_dict(response.result, response.metadata))
        metadata_values = self._prepare_metadata_response(response.metadata)
        return FeatureVector(
            names=list(features.keys()),
            values=list(features.values()),
            effective_times=[metadata_values["effective_time"].get(name) for name in features.keys()],
            slo_info=metadata_values["slo_info"],
        )

    def _prepare_metadata_response(self, metadata: Metadata) -> Dict[str, dict]:
        metadata_values = {}
        metadata_values["slo_info"] = MessageToDict(metadata.slo_info)

        times = {}
        for i, feature in enumerate(metadata.features):
            time = metadata.features[i].effective_time
            time = datetime.utcfromtimestamp(time.seconds)
            times[metadata.features[i].name] = time

        metadata_values["effective_time"] = times
        return metadata_values

    def _feature_dict(self, result: GetFeaturesResult, metadata: Metadata) -> Dict[str, Union[int, str, float, list]]:
        values = {}
        for i, feature in enumerate(result.features):
            values[metadata.features[i].name] = self._pb_to_python_value(feature, metadata.features[i].data_type)

        for i, jk in enumerate(result.join_keys):
            values[metadata.join_keys[i].name] = self._pb_to_python_value(jk, metadata.join_keys[i].data_type)
        return values

    def _prepare_request_params(self, params, join_keys, request_context_map=None, request_context_schema=None):
        request_context = request_context_map or {}

        # always returning all the metadata
        params.metadata_options.include_names = True
        params.metadata_options.include_data_types = True
        params.metadata_options.include_effective_times = True
        params.metadata_options.include_slo_info = True

        if self.feature_service_name is not None:
            params.feature_service_name = self.feature_service_name
        elif self.feature_view_name is not None:
            params.feature_view_name = self.feature_view_name
        params.workspace_name = self.workspace_name

        for k, v in join_keys.items():
            if type(v) not in (int, np.int_, str, bytes):
                raise errors.INVALID_INDIVIDUAL_JOIN_KEY_TYPE(k, type(v))
            self._python_to_pb_value(k, params.join_key_map[k], v)

        for k, v in request_context.items():
            data_type = request_context_schema.schema.get(k, None)
            # Validate request context key
            if data_type is None:
                raise errors.UNKNOWN_REQUEST_CONTEXT_KEY(sorted(request_context_schema.schema.keys()), k)
            self._request_context_to_pb_value(k, params.request_context_map[k], v, data_type)

    def _detailed_http_raise_for_status(self, http_response):
        try:
            http_response.raise_for_status()
        except HTTPError as e:
            try:
                details = http_response.json()
            except JSONDecodeError as json_e:
                raise errors.FS_BACKEND_ERROR(f"unable to process response ({http_response.status_code} error)")

            # Include the actual error message details in the exception
            # if available.
            if "message" in details and "code" in details:
                raise errors.FS_BACKEND_ERROR(details["message"])
            else:
                # Otherwise just throw the original error.
                raise e

    def _query_response_to_pandas(
        self, response: QueryFeaturesResponse, join_keys: Mapping[str, Union[int, np.int_, str, bytes]]
    ):
        response_count = len(response.results)
        data = {key: [value] * response_count for key, value in join_keys.items()}
        for result in response.results:
            features = self._feature_dict(result, response.metadata)
            # note that int(1) = np.int_(1) so dict lookup works here
            for k, v in features.items():
                if k not in data.keys():
                    data[k] = []
                data[k] = list(data[k]) + [v]  # type: ignore
        return pd.DataFrame(data=data)

    def _pb_to_python_value(self, v: Value, data_type: FeatureServerComplexDataType):
        """Converts a "Value" wrapped value into the type indicated by "type"."""
        which = v.WhichOneof("kind")
        if which is None or which == TYPE_NULL_VALUE:
            return None
        val = getattr(v, which)

        if data_type.type in (FeatureServerDataType.string, FeatureServerDataType.boolean):
            return val
        elif data_type.type in (FeatureServerDataType.float64, FeatureServerDataType.float32):
            return float(val)
        elif data_type.type == FeatureServerDataType.int64:
            # The feature server returns int64s as strings, which need to be cast.
            return int(val)
        elif data_type.type == FeatureServerDataType.array:
            return [self._pb_to_python_value(vi, data_type.element_type) for vi in val.values]
        elif data_type.type == FeatureServerDataType.struct:
            # Structs are returned as a list of values with the same order and length as the metadata's
            # `FeatureServerComplexDataType.fields`.
            struct = {}
            for i, field in enumerate(data_type.fields):
                python_value = self._pb_to_python_value(val.values[i], field.data_type)
                if python_value is not None:
                    struct[field.name] = python_value
            return struct
        else:
            raise Exception(
                f"Unexpected type '{data_type}' - Expected float64, int64, string, boolean, array, or struct."
            )

    def _python_to_pb_value(self, key: str, api_value: "Value", python_value: Union[int, np.int_, str, bool]):
        """Converts a single value from a python type to a protobuf wrapped value. Infer based on python type."""

        # NB. bool is a subclass of int in Python for some weird reason so we check for it first
        if isinstance(python_value, bool):
            api_value.bool_value = python_value
        elif isinstance(python_value, int):
            api_value.string_value = str(python_value)
        elif isinstance(python_value, np.int_):
            api_value.string_value = str(python_value)
        elif isinstance(python_value, str):
            api_value.string_value = python_value
        else:
            raise NotImplementedError(
                f"Found type '{type(python_value).__name__} for {key}' - Expected one of int, str, bool"
            )

    def _request_context_to_pb_value(
        self,
        key: str,
        api_value: "Value",
        python_value: Union[int, np.int_, str, bool, float, list],
        data_type: data_types.DataType,
    ):
        if python_value is None:
            api_value.null_value = True
        elif data_type == data_types.BoolType():
            if not isinstance(python_value, bool):
                raise TypeError(f"Invalid type for {key}: expected bool, got {type(python_value).__name__}")
            api_value.bool_value = python_value
        elif data_type == data_types.Int64Type():
            if not isinstance(python_value, (int, np.int_)):
                raise TypeError(
                    f"Invalid type for {key}: expected int or numpy.int_, got {type(python_value).__name__}"
                )
            api_value.string_value = str(python_value)
        elif data_type == data_types.StringType():
            if not isinstance(python_value, str):
                raise TypeError(f"Invalid type for {key}: expected str, got {type(python_value).__name__}")
            api_value.string_value = python_value
        elif data_type in (data_types.Float32Type(), data_types.Float64Type()):
            if not isinstance(python_value, (int, float, np.int_)):
                raise TypeError(f"Invalid type for {key}: expected int or float, got {type(python_value).__name__}")
            api_value.number_value = python_value
        elif isinstance(data_type, data_types.ArrayType):
            if not isinstance(python_value, list):
                raise TypeError(f"Invalid type for {key}: expected list, got {type(python_value).__name__}")
            api_value.list_value.SetInParent()  # Needed in the case of empty arrays.
            for item in python_value:
                list_value = api_value.list_value.values.add()
                self._request_context_to_pb_value(key + ".elementType", list_value, item, data_type.element_type)
        else:
            # should never happen
            raise NotImplementedError(f"Data type {data_type} not supported")



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/auth/authorization_service.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.api import annotations_pb2 as google_dot_api_dot_annotations__pb2
from tecton_proto.auth import principal_pb2 as tecton__proto_dot_auth_dot_principal__pb2
from tecton_proto.auth import resource_pb2 as tecton__proto_dot_auth_dot_resource__pb2
from tecton_proto.auth import service_pb2 as tecton__proto_dot_auth_dot_service__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n-tecton_proto/auth/authorization_service.proto\x12\x11tecton_proto.auth\x1a\x1cgoogle/api/annotations.proto\x1a!tecton_proto/auth/principal.proto\x1a tecton_proto/auth/resource.proto\x1a\x1ftecton_proto/auth/service.proto\"\x11\n\x0fGetRolesRequest\"K\n\x10GetRolesResponse\x12\x37\n\x05roles\x18\x01 \x03(\x0b\x32!.tecton_proto.auth.RoleDefinitionR\x05roles\"\x85\x03\n\x0eRoleDefinition\x12\x0e\n\x02id\x18\x01 \x01(\tR\x02id\x12\x12\n\x04name\x18\x02 \x01(\tR\x04name\x12 \n\x0b\x64\x65scription\x18\x03 \x01(\tR\x0b\x64\x65scription\x12`\n\x1c\x61ssignable_on_resource_types\x18\x04 \x03(\x0e\x32\x1f.tecton_proto.auth.ResourceTypeR\x19\x61ssignableOnResourceTypes\x12\x63\n\x1d\x61ssignable_to_principal_types\x18\x05 \x03(\x0e\x32 .tecton_proto.auth.PrincipalTypeR\x1a\x61ssignableToPrincipalTypes\x12I\n\x0bpermissions\x18\x06 \x03(\x0b\x32\'.tecton_proto.auth.PermissionDefinitionR\x0bpermissions\x12\x1b\n\tlegacy_id\x18\x07 \x01(\tR\x08legacyId\"m\n\x14PermissionDefinition\x12\x0e\n\x02id\x18\x01 \x01(\tR\x02id\x12 \n\x0b\x64\x65scription\x18\x02 \x01(\tR\x0b\x64\x65scription\x12#\n\ris_authorized\x18\x03 \x01(\x08R\x0cisAuthorized\"\x84\x02\n\x17GetAssignedRolesRequest\x12G\n\x0eprincipal_type\x18\x01 \x01(\x0e\x32 .tecton_proto.auth.PrincipalTypeR\rprincipalType\x12!\n\x0cprincipal_id\x18\x02 \x01(\tR\x0bprincipalId\x12\x44\n\rresource_type\x18\x03 \x01(\x0e\x32\x1f.tecton_proto.auth.ResourceTypeR\x0cresourceType\x12!\n\x0cresource_ids\x18\x04 \x03(\tR\x0bresourceIds\x12\x14\n\x05roles\x18\x05 \x03(\tR\x05roles\"^\n\x18GetAssignedRolesResponse\x12\x42\n\x0b\x61ssignments\x18\x01 \x03(\x0b\x32 .tecton_proto.auth.ResourceRolesR\x0b\x61ssignments\"\xc5\x01\n\x16GetIsAuthorizedRequest\x12G\n\x0eprincipal_type\x18\x01 \x01(\x0e\x32 .tecton_proto.auth.PrincipalTypeR\rprincipalType\x12!\n\x0cprincipal_id\x18\x02 \x01(\tR\x0bprincipalId\x12?\n\x0bpermissions\x18\x03 \x03(\x0b\x32\x1d.tecton_proto.auth.PermissionR\x0bpermissions\"Z\n\x17GetIsAuthorizedResponse\x12?\n\x0bpermissions\x18\x01 \x03(\x0b\x32\x1d.tecton_proto.auth.PermissionR\x0bpermissions\"\x8c\x01\n\rResourceRoles\x12\x44\n\rresource_type\x18\x03 \x01(\x0e\x32\x1f.tecton_proto.auth.ResourceTypeR\x0cresourceType\x12\x1f\n\x0bresource_id\x18\x01 \x01(\tR\nresourceId\x12\x14\n\x05roles\x18\x02 \x03(\tR\x05roles\"U\n\x12\x41ssignRolesRequest\x12?\n\x0b\x61ssignments\x18\x01 \x03(\x0b\x32\x1d.tecton_proto.auth.AssignmentR\x0b\x61ssignments\"\x15\n\x13\x41ssignRolesResponse\"W\n\x14UnassignRolesRequest\x12?\n\x0b\x61ssignments\x18\x01 \x03(\x0b\x32\x1d.tecton_proto.auth.AssignmentR\x0b\x61ssignments\"\x17\n\x15UnassignRolesResponse\"\xf3\x01\n\nAssignment\x12\x44\n\rresource_type\x18\x01 \x01(\x0e\x32\x1f.tecton_proto.auth.ResourceTypeR\x0cresourceType\x12\x1f\n\x0bresource_id\x18\x02 \x01(\tR\nresourceId\x12\x12\n\x04role\x18\x03 \x01(\tR\x04role\x12G\n\x0eprincipal_type\x18\x04 \x01(\x0e\x32 .tecton_proto.auth.PrincipalTypeR\rprincipalType\x12!\n\x0cprincipal_id\x18\x05 \x01(\tR\x0bprincipalId\"\x80\x02\n\x15\x41ssignRolesPutRequest\x12\x44\n\rresource_type\x18\x01 \x01(\x0e\x32\x1f.tecton_proto.auth.ResourceTypeR\x0cresourceType\x12\x1f\n\x0bresource_id\x18\x02 \x01(\tR\nresourceId\x12G\n\x0eprincipal_type\x18\x03 \x01(\x0e\x32 .tecton_proto.auth.PrincipalTypeR\rprincipalType\x12!\n\x0cprincipal_id\x18\x04 \x01(\tR\x0bprincipalId\x12\x14\n\x05roles\x18\x05 \x03(\tR\x05roles\"\x18\n\x16\x41ssignRolesPutResponse\"\xe9\x01\n\x1dGetAuthorizedResourcesRequest\x12G\n\x0eprincipal_type\x18\x01 \x01(\x0e\x32 .tecton_proto.auth.PrincipalTypeR\rprincipalType\x12!\n\x0cprincipal_id\x18\x02 \x01(\tR\x0bprincipalId\x12\x44\n\rresource_type\x18\x03 \x01(\x0e\x32\x1f.tecton_proto.auth.ResourceTypeR\x0cresourceType\x12\x16\n\x06\x61\x63tion\x18\x04 \x01(\tR\x06\x61\x63tion\"{\n\x1eGetAuthorizedResourcesResponse\x12Y\n\x14\x61uthorized_resources\x18\x01 \x03(\x0b\x32&.tecton_proto.auth.AuthorizedResourcesR\x13\x61uthorizedResources\"P\n\x13\x41uthorizedResources\x12\x1f\n\x0bresource_id\x18\x01 \x01(\tR\nresourceId\x12\x18\n\x07\x61\x63tions\x18\x02 \x03(\tR\x07\x61\x63tions\"\xe6\x01\n\x1cGetAssignedPrincipalsRequest\x12\x44\n\rresource_type\x18\x01 \x01(\x0e\x32\x1f.tecton_proto.auth.ResourceTypeR\x0cresourceType\x12\x1f\n\x0bresource_id\x18\x02 \x01(\tR\nresourceId\x12\x14\n\x05roles\x18\x03 \x03(\tR\x05roles\x12I\n\x0fprincipal_types\x18\x04 \x03(\x0e\x32 .tecton_proto.auth.PrincipalTypeR\x0eprincipalTypes\"e\n\x1dGetAssignedPrincipalsResponse\x12\x44\n\x0b\x61ssignments\x18\x01 \x03(\x0b\x32\".tecton_proto.auth.AssignmentBasicR\x0b\x61ssignments\"h\n\x0f\x41ssignmentBasic\x12?\n\tprincipal\x18\x01 \x01(\x0b\x32!.tecton_proto.auth.PrincipalBasicR\tprincipal\x12\x14\n\x05roles\x18\x02 \x03(\tR\x05roles\"\xee\x01\n\x18GetAppPermissionsRequest\x12G\n\x0eprincipal_type\x18\x01 \x01(\x0e\x32 .tecton_proto.auth.PrincipalTypeR\rprincipalType\x12!\n\x0cprincipal_id\x18\x02 \x01(\tR\x0bprincipalId\x12\x66\n\x19resource_type_permissions\x18\x05 \x03(\x0b\x32*.tecton_proto.auth.ResourceTypePermissionsR\x17resourceTypePermissions\"y\n\x17ResourceTypePermissions\x12\x44\n\rresource_type\x18\x01 \x01(\x0e\x32\x1f.tecton_proto.auth.ResourceTypeR\x0cresourceType\x12\x18\n\x07\x61\x63tions\x18\x02 \x03(\tR\x07\x61\x63tions\"\x93\x01\n\x19GetAppPermissionsResponse\x12v\n\x1fresource_type_permission_values\x18\x01 \x03(\x0b\x32/.tecton_proto.auth.ResourceTypePermissionValuesR\x1cresourceTypePermissionValues\"\xb5\x01\n\x1cResourceTypePermissionValues\x12\x44\n\rresource_type\x18\x01 \x01(\x0e\x32\x1f.tecton_proto.auth.ResourceTypeR\x0cresourceType\x12O\n\x11permission_values\x18\x02 \x03(\x0b\x32\".tecton_proto.auth.PermissionValueR\x10permissionValues\"\x8b\x01\n\nPermission\x12\x44\n\rresource_type\x18\x01 \x01(\x0e\x32\x1f.tecton_proto.auth.ResourceTypeR\x0cresourceType\x12\x1f\n\x0bresource_id\x18\x02 \x01(\tR\nresourceId\x12\x16\n\x06\x61\x63tion\x18\x03 \x01(\tR\x06\x61\x63tion\"N\n\x0fPermissionValue\x12\x16\n\x06\x61\x63tion\x18\x01 \x01(\tR\x06\x61\x63tion\x12#\n\ris_authorized\x18\x02 \x01(\x08R\x0cisAuthorized\"\x97\x02\n\x1eGetWorkspacePermissionsRequest\x12G\n\x0eprincipal_type\x18\x01 \x01(\x0e\x32 .tecton_proto.auth.PrincipalTypeR\rprincipalType\x12!\n\x0cprincipal_id\x18\x02 \x01(\tR\x0bprincipalId\x12!\n\x0cworkspace_id\x18\x03 \x01(\tR\x0bworkspaceId\x12\x66\n\x19resource_type_permissions\x18\x04 \x03(\x0b\x32*.tecton_proto.auth.ResourceTypePermissionsR\x17resourceTypePermissions\"\x99\x01\n\x1fGetWorkspacePermissionsResponse\x12v\n\x1fresource_type_permission_values\x18\x01 \x03(\x0b\x32/.tecton_proto.auth.ResourceTypePermissionValuesR\x1cresourceTypePermissionValues2\xe5\r\n\x14\x41uthorizationService\x12\x8b\x01\n\x08GetRoles\x12\".tecton_proto.auth.GetRolesRequest\x1a#.tecton_proto.auth.GetRolesResponse\"6\x82\xd3\xe4\x93\x02(\"#/v1/authorization-service/get-roles:\x01*\xa2\xbc\xe6\xc0\x05\x02\x10\x01\x12\xac\x01\n\x10GetAssignedRoles\x12*.tecton_proto.auth.GetAssignedRolesRequest\x1a+.tecton_proto.auth.GetAssignedRolesResponse\"?\x82\xd3\xe4\x93\x02\x31\",/v1/authorization-service/get-assigned-roles:\x01*\xa2\xbc\xe6\xc0\x05\x02@\x01\x12\x97\x01\n\x0b\x41ssignRoles\x12%.tecton_proto.auth.AssignRolesRequest\x1a&.tecton_proto.auth.AssignRolesResponse\"9\x82\xd3\xe4\x93\x02+\"&/v1/authorization-service/assign-roles:\x01*\xa2\xbc\xe6\xc0\x05\x02@\x01\x12\x9f\x01\n\rUnassignRoles\x12\'.tecton_proto.auth.UnassignRolesRequest\x1a(.tecton_proto.auth.UnassignRolesResponse\";\x82\xd3\xe4\x93\x02-\"(/v1/authorization-service/unassign-roles:\x01*\xa2\xbc\xe6\xc0\x05\x02@\x01\x12\xa0\x01\n\x0e\x41ssignRolesPut\x12(.tecton_proto.auth.AssignRolesPutRequest\x1a).tecton_proto.auth.AssignRolesPutResponse\"9\x82\xd3\xe4\x93\x02+\x1a&/v1/authorization-service/assign-roles:\x01*\xa2\xbc\xe6\xc0\x05\x02@\x01\x12\xc4\x01\n\x16GetAuthorizedResources\x12\x30.tecton_proto.auth.GetAuthorizedResourcesRequest\x1a\x31.tecton_proto.auth.GetAuthorizedResourcesResponse\"E\x82\xd3\xe4\x93\x02\x37\"2/v1/authorization-service/get-authorized-resources:\x01*\xa2\xbc\xe6\xc0\x05\x02@\x01\x12\xc0\x01\n\x15GetAssignedPrincipals\x12/.tecton_proto.auth.GetAssignedPrincipalsRequest\x1a\x30.tecton_proto.auth.GetAssignedPrincipalsResponse\"D\x82\xd3\xe4\x93\x02\x36\"1/v1/authorization-service/get-assigned-principals:\x01*\xa2\xbc\xe6\xc0\x05\x02@\x01\x12\xb0\x01\n\x11GetAppPermissions\x12+.tecton_proto.auth.GetAppPermissionsRequest\x1a,.tecton_proto.auth.GetAppPermissionsResponse\"@\x82\xd3\xe4\x93\x02\x32\"-/v1/authorization-service/get-app-permissions:\x01*\xa2\xbc\xe6\xc0\x05\x02@\x01\x12\xc8\x01\n\x17GetWorkspacePermissions\x12\x31.tecton_proto.auth.GetWorkspacePermissionsRequest\x1a\x32.tecton_proto.auth.GetWorkspacePermissionsResponse\"F\x82\xd3\xe4\x93\x02\x38\"3/v1/authorization-service/get-workspace-permissions:\x01*\xa2\xbc\xe6\xc0\x05\x02@\x01\x12\xa8\x01\n\x0fGetIsAuthorized\x12).tecton_proto.auth.GetIsAuthorizedRequest\x1a*.tecton_proto.auth.GetIsAuthorizedResponse\">\x82\xd3\xe4\x93\x02\x30\"+/v1/authorization-service/get-is-authorized:\x01*\xa2\xbc\xe6\xc0\x05\x02@\x01\x42V\n\x0f\x63om.tecton.authB\x19\x41uthorizationServiceProtoP\x01Z&github.com/tecton-ai/tecton_proto/auth')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tecton_proto.auth.authorization_service_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\017com.tecton.authB\031AuthorizationServiceProtoP\001Z&github.com/tecton-ai/tecton_proto/auth'
  _AUTHORIZATIONSERVICE.methods_by_name['GetRoles']._options = None
  _AUTHORIZATIONSERVICE.methods_by_name['GetRoles']._serialized_options = b'\202\323\344\223\002(\"#/v1/authorization-service/get-roles:\001*\242\274\346\300\005\002\020\001'
  _AUTHORIZATIONSERVICE.methods_by_name['GetAssignedRoles']._options = None
  _AUTHORIZATIONSERVICE.methods_by_name['GetAssignedRoles']._serialized_options = b'\202\323\344\223\0021\",/v1/authorization-service/get-assigned-roles:\001*\242\274\346\300\005\002@\001'
  _AUTHORIZATIONSERVICE.methods_by_name['AssignRoles']._options = None
  _AUTHORIZATIONSERVICE.methods_by_name['AssignRoles']._serialized_options = b'\202\323\344\223\002+\"&/v1/authorization-service/assign-roles:\001*\242\274\346\300\005\002@\001'
  _AUTHORIZATIONSERVICE.methods_by_name['UnassignRoles']._options = None
  _AUTHORIZATIONSERVICE.methods_by_name['UnassignRoles']._serialized_options = b'\202\323\344\223\002-\"(/v1/authorization-service/unassign-roles:\001*\242\274\346\300\005\002@\001'
  _AUTHORIZATIONSERVICE.methods_by_name['AssignRolesPut']._options = None
  _AUTHORIZATIONSERVICE.methods_by_name['AssignRolesPut']._serialized_options = b'\202\323\344\223\002+\032&/v1/authorization-service/assign-roles:\001*\242\274\346\300\005\002@\001'
  _AUTHORIZATIONSERVICE.methods_by_name['GetAuthorizedResources']._options = None
  _AUTHORIZATIONSERVICE.methods_by_name['GetAuthorizedResources']._serialized_options = b'\202\323\344\223\0027\"2/v1/authorization-service/get-authorized-resources:\001*\242\274\346\300\005\002@\001'
  _AUTHORIZATIONSERVICE.methods_by_name['GetAssignedPrincipals']._options = None
  _AUTHORIZATIONSERVICE.methods_by_name['GetAssignedPrincipals']._serialized_options = b'\202\323\344\223\0026\"1/v1/authorization-service/get-assigned-principals:\001*\242\274\346\300\005\002@\001'
  _AUTHORIZATIONSERVICE.methods_by_name['GetAppPermissions']._options = None
  _AUTHORIZATIONSERVICE.methods_by_name['GetAppPermissions']._serialized_options = b'\202\323\344\223\0022\"-/v1/authorization-service/get-app-permissions:\001*\242\274\346\300\005\002@\001'
  _AUTHORIZATIONSERVICE.methods_by_name['GetWorkspacePermissions']._options = None
  _AUTHORIZATIONSERVICE.methods_by_name['GetWorkspacePermissions']._serialized_options = b'\202\323\344\223\0028\"3/v1/authorization-service/get-workspace-permissions:\001*\242\274\346\300\005\002@\001'
  _AUTHORIZATIONSERVICE.methods_by_name['GetIsAuthorized']._options = None
  _AUTHORIZATIONSERVICE.methods_by_name['GetIsAuthorized']._serialized_options = b'\202\323\344\223\0020\"+/v1/authorization-service/get-is-authorized:\001*\242\274\346\300\005\002@\001'
  _GETROLESREQUEST._serialized_start=200
  _GETROLESREQUEST._serialized_end=217
  _GETROLESRESPONSE._serialized_start=219
  _GETROLESRESPONSE._serialized_end=294
  _ROLEDEFINITION._serialized_start=297
  _ROLEDEFINITION._serialized_end=686
  _PERMISSIONDEFINITION._serialized_start=688
  _PERMISSIONDEFINITION._serialized_end=797
  _GETASSIGNEDROLESREQUEST._serialized_start=800
  _GETASSIGNEDROLESREQUEST._serialized_end=1060
  _GETASSIGNEDROLESRESPONSE._serialized_start=1062
  _GETASSIGNEDROLESRESPONSE._serialized_end=1156
  _GETISAUTHORIZEDREQUEST._serialized_start=1159
  _GETISAUTHORIZEDREQUEST._serialized_end=1356
  _GETISAUTHORIZEDRESPONSE._serialized_start=1358
  _GETISAUTHORIZEDRESPONSE._serialized_end=1448
  _RESOURCEROLES._serialized_start=1451
  _RESOURCEROLES._serialized_end=1591
  _ASSIGNROLESREQUEST._serialized_start=1593
  _ASSIGNROLESREQUEST._serialized_end=1678
  _ASSIGNROLESRESPONSE._serialized_start=1680
  _ASSIGNROLESRESPONSE._serialized_end=1701
  _UNASSIGNROLESREQUEST._serialized_start=1703
  _UNASSIGNROLESREQUEST._serialized_end=1790
  _UNASSIGNROLESRESPONSE._serialized_start=1792
  _UNASSIGNROLESRESPONSE._serialized_end=1815
  _ASSIGNMENT._serialized_start=1818
  _ASSIGNMENT._serialized_end=2061
  _ASSIGNROLESPUTREQUEST._serialized_start=2064
  _ASSIGNROLESPUTREQUEST._serialized_end=2320
  _ASSIGNROLESPUTRESPONSE._serialized_start=2322
  _ASSIGNROLESPUTRESPONSE._serialized_end=2346
  _GETAUTHORIZEDRESOURCESREQUEST._serialized_start=2349
  _GETAUTHORIZEDRESOURCESREQUEST._serialized_end=2582
  _GETAUTHORIZEDRESOURCESRESPONSE._serialized_start=2584
  _GETAUTHORIZEDRESOURCESRESPONSE._serialized_end=2707
  _AUTHORIZEDRESOURCES._serialized_start=2709
  _AUTHORIZEDRESOURCES._serialized_end=2789
  _GETASSIGNEDPRINCIPALSREQUEST._serialized_start=2792
  _GETASSIGNEDPRINCIPALSREQUEST._serialized_end=3022
  _GETASSIGNEDPRINCIPALSRESPONSE._serialized_start=3024
  _GETASSIGNEDPRINCIPALSRESPONSE._serialized_end=3125
  _ASSIGNMENTBASIC._serialized_start=3127
  _ASSIGNMENTBASIC._serialized_end=3231
  _GETAPPPERMISSIONSREQUEST._serialized_start=3234
  _GETAPPPERMISSIONSREQUEST._serialized_end=3472
  _RESOURCETYPEPERMISSIONS._serialized_start=3474
  _RESOURCETYPEPERMISSIONS._serialized_end=3595
  _GETAPPPERMISSIONSRESPONSE._serialized_start=3598
  _GETAPPPERMISSIONSRESPONSE._serialized_end=3745
  _RESOURCETYPEPERMISSIONVALUES._serialized_start=3748
  _RESOURCETYPEPERMISSIONVALUES._serialized_end=3929
  _PERMISSION._serialized_start=3932
  _PERMISSION._serialized_end=4071
  _PERMISSIONVALUE._serialized_start=4073
  _PERMISSIONVALUE._serialized_end=4151
  _GETWORKSPACEPERMISSIONSREQUEST._serialized_start=4154
  _GETWORKSPACEPERMISSIONSREQUEST._serialized_end=4433
  _GETWORKSPACEPERMISSIONSRESPONSE._serialized_start=4436
  _GETWORKSPACEPERMISSIONSRESPONSE._serialized_end=4589
  _AUTHORIZATIONSERVICE._serialized_start=4592
  _AUTHORIZATIONSERVICE._serialized_end=6357
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/databricks_api/workspace.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n+tecton_proto/databricks_api/workspace.proto\x12\x1btecton_proto.databricks_api\"*\n\x14ListWorkspaceRequest\x12\x12\n\x04path\x18\x01 \x01(\tR\x04path\"Z\n\x15ListWorkspaceResponse\x12\x41\n\x07objects\x18\x01 \x03(\x0b\x32\'.tecton_proto.databricks_api.ObjectInfoR\x07objects\"z\n\nObjectInfo\x12\x1f\n\x0bobject_type\x18\x01 \x01(\tR\nobjectType\x12\x1b\n\tobject_id\x18\x02 \x01(\x03R\x08objectId\x12\x12\n\x04path\x18\x03 \x01(\tR\x04path\x12\x1a\n\x08language\x18\x04 \x01(\tR\x08languageB\x1d\n\x19\x63om.tecton.databricks_apiP\x01')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tecton_proto.databricks_api.workspace_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\031com.tecton.databricks_apiP\001'
  _LISTWORKSPACEREQUEST._serialized_start=76
  _LISTWORKSPACEREQUEST._serialized_end=118
  _LISTWORKSPACERESPONSE._serialized_start=120
  _LISTWORKSPACERESPONSE._serialized_end=210
  _OBJECTINFO._serialized_start=212
  _OBJECTINFO._serialized_end=334
# @@protoc_insertion_point(module_scope)



from dataclasses import dataclass
from datetime import datetime
from typing import Optional


@dataclass
class FilterContext:
    """
    ``FilterContext`` is passed as an argument to Data Source Function for time filtering.
    """

    start_time: Optional[datetime]
    end_time: Optional[datetime]



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/common/fco_locator.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from tecton_proto.common import id_pb2 as tecton__proto_dot_common_dot_id__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n%tecton_proto/common/fco_locator.proto\x12\x13tecton_proto.common\x1a\x1ctecton_proto/common/id.proto\"v\n\nFcoLocator\x12\x14\n\x04name\x18\x01 \x01(\tH\x00R\x04name\x12)\n\x02id\x18\x02 \x01(\x0b\x32\x17.tecton_proto.common.IdH\x00R\x02id\x12\x1c\n\tworkspace\x18\x03 \x01(\tR\tworkspaceB\t\n\x07locator\"[\n\x0cIdFcoLocator\x12\'\n\x02id\x18\x02 \x01(\x0b\x32\x17.tecton_proto.common.IdR\x02id\x12\x1c\n\tworkspace\x18\x03 \x01(\tR\tworkspaceJ\x04\x08\x01\x10\x02\x42\x15\n\x11\x63om.tecton.commonP\x01')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tecton_proto.common.fco_locator_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\021com.tecton.commonP\001'
  _FCOLOCATOR._serialized_start=92
  _FCOLOCATOR._serialized_end=210
  _IDFCOLOCATOR._serialized_start=212
  _IDFCOLOCATOR._serialized_end=303
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/data/state_update.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from tecton_proto.common import id_pb2 as tecton__proto_dot_common_dot_id__pb2
from tecton_proto.args import fco_args_pb2 as tecton__proto_dot_args_dot_fco__args__pb2
from tecton_proto.args import repo_metadata_pb2 as tecton__proto_dot_args_dot_repo__metadata__pb2
from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2
from google.protobuf import duration_pb2 as google_dot_protobuf_dot_duration__pb2
from tecton_proto.args import diff_options_pb2 as tecton__proto_dot_args_dot_diff__options__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n$tecton_proto/data/state_update.proto\x12\x11tecton_proto.data\x1a\x1ctecton_proto/common/id.proto\x1a tecton_proto/args/fco_args.proto\x1a%tecton_proto/args/repo_metadata.proto\x1a\x1fgoogle/protobuf/timestamp.proto\x1a\x1egoogle/protobuf/duration.proto\x1a$tecton_proto/args/diff_options.proto\"\xd0\x01\n\x0f\x46\x63oPropertyDiff\x12#\n\rproperty_name\x18\x01 \x01(\tR\x0cpropertyName\x12!\n\x0cval_existing\x18\x02 \x01(\tR\x0bvalExisting\x12!\n\x0cval_declared\x18\x03 \x01(\tR\x0bvalDeclared\x12R\n\x0erendering_type\x18\x04 \x01(\x0e\x32+.tecton_proto.args.FcoPropertyRenderingTypeR\rrenderingType\"\x87\x01\n\x18\x46\x63oTransitionSideEffects\x12k\n\x13stream_restart_type\x18\x01 \x01(\x0e\x32;.tecton_proto.data.FcoTransitionSideEffectStreamRestartTypeR\x11streamRestartType\"\xe6\x01\n\x1c\x42\x61tchMaterializationTaskDiff\x12%\n\x0e\x64isplay_string\x18\x01 \x01(\tR\rdisplayString\x12\x46\n\x11schedule_interval\x18\x02 \x01(\x0b\x32\x19.google.protobuf.DurationR\x10scheduleInterval\x12W\n\x0b\x64\x65stination\x18\x03 \x01(\x0e\x32\x35.tecton_proto.data.MaterializationTaskDiffDestinationR\x0b\x64\x65stination\"F\n\x1dStreamMaterializationTaskDiff\x12%\n\x0e\x64isplay_string\x18\x01 \x01(\tR\rdisplayString\"\xd7\x02\n\x1f\x42\x61\x63kfillMaterializationTaskDiff\x12%\n\x0e\x64isplay_string\x18\x01 \x01(\tR\rdisplayString\x12H\n\x12\x66\x65\x61ture_start_time\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\x10\x66\x65\x61tureStartTime\x12\x44\n\x10\x66\x65\x61ture_end_time\x18\x03 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\x0e\x66\x65\x61tureEndTime\x12$\n\x0enumber_of_jobs\x18\x04 \x01(\x05R\x0cnumberOfJobs\x12W\n\x0b\x64\x65stination\x18\x05 \x01(\x0e\x32\x35.tecton_proto.data.MaterializationTaskDiffDestinationR\x0b\x64\x65stination\"\xae\x02\n\x13MaterializationInfo\x12\x62\n\x13\x62\x61\x63kfill_task_diffs\x18\x01 \x03(\x0b\x32\x32.tecton_proto.data.BackfillMaterializationTaskDiffR\x11\x62\x61\x63kfillTaskDiffs\x12W\n\x0f\x62\x61tch_task_diff\x18\x02 \x01(\x0b\x32/.tecton_proto.data.BatchMaterializationTaskDiffR\rbatchTaskDiff\x12Z\n\x10stream_task_diff\x18\x03 \x01(\x0b\x32\x30.tecton_proto.data.StreamMaterializationTaskDiffR\x0estreamTaskDiff\"\xc3\x03\n\x07\x46\x63oDiff\x12\x38\n\x04type\x18\x01 \x01(\x0e\x32$.tecton_proto.data.FcoTransitionTypeR\x04type\x12\x63\n\x17transition_side_effects\x18\x0c \x01(\x0b\x32+.tecton_proto.data.FcoTransitionSideEffectsR\x15transitionSideEffects\x12?\n\rexisting_args\x18\x07 \x01(\x0b\x32\x1a.tecton_proto.args.FcoArgsR\x0c\x65xistingArgs\x12?\n\rdeclared_args\x18\x08 \x01(\x0b\x32\x1a.tecton_proto.args.FcoArgsR\x0c\x64\x65\x63laredArgs\x12\x36\n\x04\x64iff\x18\x04 \x03(\x0b\x32\".tecton_proto.data.FcoPropertyDiffR\x04\x64iff\x12Y\n\x14materialization_info\x18\x0b \x01(\x0b\x32&.tecton_proto.data.MaterializationInfoR\x13materializationInfoJ\x04\x08\t\x10\n\"C\n\x0b\x46\x63oFieldRef\x12.\n\x06\x66\x63o_id\x18\x01 \x01(\x0b\x32\x17.tecton_proto.common.IdR\x05\x66\x63oIdJ\x04\x08\x02\x10\x03\"h\n\x11ValidationMessage\x12\x18\n\x07message\x18\x01 \x01(\tR\x07message\x12\x39\n\x08\x66\x63o_refs\x18\x02 \x03(\x0b\x32\x1e.tecton_proto.data.FcoFieldRefR\x07\x66\x63oRefs\"\x92\x01\n\x10ValidationResult\x12<\n\x06\x65rrors\x18\x01 \x03(\x0b\x32$.tecton_proto.data.ValidationMessageR\x06\x65rrors\x12@\n\x08warnings\x18\x02 \x03(\x0b\x32$.tecton_proto.data.ValidationMessageR\x08warnings\"\xd0\x01\n\x14SuccessfulPlanOutput\x12#\n\rstring_output\x18\x01 \x01(\tR\x0cstringOutput\x12\x1f\n\x0bjson_output\x18\x02 \x01(\tR\njsonOutput\x12%\n\x0e\x61pply_warnings\x18\x03 \x01(\tR\rapplyWarnings\x12(\n\x10num_fcos_changed\x18\x04 \x01(\x05R\x0enumFcosChanged\x12!\n\x0cnum_warnings\x18\x05 \x01(\x05R\x0bnumWarnings\"\xe3\x02\n\x12StateUpdateRequest\x12\x1c\n\tworkspace\x18\x01 \x01(\tR\tworkspace\x12\x35\n\x08\x66\x63o_args\x18\x02 \x03(\x0b\x32\x1a.tecton_proto.args.FcoArgsR\x07\x66\x63oArgs\x12R\n\x10repo_source_info\x18\x03 \x01(\x0b\x32(.tecton_proto.args.FeatureRepoSourceInfoR\x0erepoSourceInfo\x12-\n\x12suppress_recreates\x18\n \x01(\x08R\x11suppressRecreates\x12\x1f\n\x0bupgrade_all\x18\x04 \x01(\x08R\nupgradeAll\x12!\n\x0crequested_by\x18\x05 \x01(\tR\x0brequestedBy\x12\x1f\n\x0bsdk_version\x18\t \x01(\tR\nsdkVersionJ\x04\x08\x06\x10\x07J\x04\x08\x07\x10\x08J\x04\x08\x08\x10\t\"\xc8\x01\n\x10StateUpdateEntry\x12\x1b\n\tcommit_id\x18\x01 \x01(\tR\x08\x63ommitId\x12\x1d\n\napplied_by\x18\x02 \x01(\tR\tappliedBy\x12\x39\n\napplied_at\x18\x03 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\tappliedAt\x12\x1c\n\tworkspace\x18\x04 \x01(\tR\tworkspace\x12\x1f\n\x0bsdk_version\x18\x05 \x01(\tR\nsdkVersion\"\xd9\x02\n\x16StateUpdatePlanSummary\x12L\n\ndiff_items\x18\x01 \x03(\x0b\x32-.tecton_proto.data.StateUpdatePlanSummaryDiffR\tdiffItems\x12\x1d\n\napplied_by\x18\x02 \x01(\tR\tappliedBy\x12\x1d\n\ncreated_by\x18\x03 \x01(\tR\tcreatedBy\x12\x39\n\napplied_at\x18\x04 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\tappliedAt\x12\x39\n\ncreated_at\x18\x05 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\tcreatedAt\x12\x1c\n\tworkspace\x18\x06 \x01(\tR\tworkspace\x12\x1f\n\x0bsdk_version\x18\x07 \x01(\tR\nsdkVersion\"\xa1\x03\n\x1aStateUpdatePlanSummaryDiff\x12\x19\n\x08\x66\x63o_type\x18\x01 \x01(\tR\x07\x66\x63oType\x12\x38\n\x04type\x18\x02 \x01(\x0e\x32$.tecton_proto.data.FcoTransitionTypeR\x04type\x12\x63\n\x17transition_side_effects\x18\x07 \x01(\x0b\x32+.tecton_proto.data.FcoTransitionSideEffectsR\x15transitionSideEffects\x12\x38\n\x05\x64iffs\x18\x03 \x03(\x0b\x32\".tecton_proto.data.FcoPropertyDiffR\x05\x64iffs\x12\x12\n\x04name\x18\x04 \x01(\tR\x04name\x12 \n\x0b\x64\x65scription\x18\x05 \x01(\tR\x0b\x64\x65scription\x12Y\n\x14materialization_info\x18\x06 \x01(\x0b\x32&.tecton_proto.data.MaterializationInfoR\x13materializationInfo*n\n\x11\x46\x63oTransitionType\x12\x0b\n\x07UNKNOWN\x10\x00\x12\n\n\x06\x43REATE\x10\x01\x12\n\n\x06\x44\x45LETE\x10\x02\x12\x0b\n\x07UPGRADE\x10\x05\x12\n\n\x06UPDATE\x10\x03\x12\x0c\n\x08RECREATE\x10\x04\x12\r\n\tUNCHANGED\x10\x06*\x95\x01\n(FcoTransitionSideEffectStreamRestartType\x12\x17\n\x13RESTART_STREAM_NONE\x10\x00\x12$\n RESTART_STREAM_REUSE_CHECKPOINTS\x10\x01\x12*\n&RESTART_STREAM_CHECKPOINTS_INVALIDATED\x10\x02*\xfe\x01\n\"MaterializationTaskDiffDestination\x12\x35\n1MATERIALIZATION_TASK_DIFF_DESTINATION_UNSPECIFIED\x10\x00\x12\x31\n-MATERIALIZATION_TASK_DIFF_DESTINATION_OFFLINE\x10\x01\x12\x30\n,MATERIALIZATION_TASK_DIFF_DESTINATION_ONLINE\x10\x02\x12<\n8MATERIALIZATION_TASK_DIFF_DESTINATION_ONLINE_AND_OFFLINE\x10\x03\x42\x13\n\x0f\x63om.tecton.dataP\x01')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tecton_proto.data.state_update_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\017com.tecton.dataP\001'
  _FCOTRANSITIONTYPE._serialized_start=3888
  _FCOTRANSITIONTYPE._serialized_end=3998
  _FCOTRANSITIONSIDEEFFECTSTREAMRESTARTTYPE._serialized_start=4001
  _FCOTRANSITIONSIDEEFFECTSTREAMRESTARTTYPE._serialized_end=4150
  _MATERIALIZATIONTASKDIFFDESTINATION._serialized_start=4153
  _MATERIALIZATIONTASKDIFFDESTINATION._serialized_end=4407
  _FCOPROPERTYDIFF._serialized_start=266
  _FCOPROPERTYDIFF._serialized_end=474
  _FCOTRANSITIONSIDEEFFECTS._serialized_start=477
  _FCOTRANSITIONSIDEEFFECTS._serialized_end=612
  _BATCHMATERIALIZATIONTASKDIFF._serialized_start=615
  _BATCHMATERIALIZATIONTASKDIFF._serialized_end=845
  _STREAMMATERIALIZATIONTASKDIFF._serialized_start=847
  _STREAMMATERIALIZATIONTASKDIFF._serialized_end=917
  _BACKFILLMATERIALIZATIONTASKDIFF._serialized_start=920
  _BACKFILLMATERIALIZATIONTASKDIFF._serialized_end=1263
  _MATERIALIZATIONINFO._serialized_start=1266
  _MATERIALIZATIONINFO._serialized_end=1568
  _FCODIFF._serialized_start=1571
  _FCODIFF._serialized_end=2022
  _FCOFIELDREF._serialized_start=2024
  _FCOFIELDREF._serialized_end=2091
  _VALIDATIONMESSAGE._serialized_start=2093
  _VALIDATIONMESSAGE._serialized_end=2197
  _VALIDATIONRESULT._serialized_start=2200
  _VALIDATIONRESULT._serialized_end=2346
  _SUCCESSFULPLANOUTPUT._serialized_start=2349
  _SUCCESSFULPLANOUTPUT._serialized_end=2557
  _STATEUPDATEREQUEST._serialized_start=2560
  _STATEUPDATEREQUEST._serialized_end=2915
  _STATEUPDATEENTRY._serialized_start=2918
  _STATEUPDATEENTRY._serialized_end=3118
  _STATEUPDATEPLANSUMMARY._serialized_start=3121
  _STATEUPDATEPLANSUMMARY._serialized_end=3466
  _STATEUPDATEPLANSUMMARYDIFF._serialized_start=3469
  _STATEUPDATEPLANSUMMARYDIFF._serialized_end=3886
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/data/freshness_status.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2
from google.protobuf import duration_pb2 as google_dot_protobuf_dot_duration__pb2
from tecton_proto.common import id_pb2 as tecton__proto_dot_common_dot_id__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n(tecton_proto/data/freshness_status.proto\x12\x11tecton_proto.data\x1a\x1fgoogle/protobuf/timestamp.proto\x1a\x1egoogle/protobuf/duration.proto\x1a\x1ctecton_proto/common/id.proto\"\xad\x03\n\x0f\x46reshnessStatus\x12?\n\x0f\x66\x65\x61ture_view_id\x18\x01 \x01(\x0b\x32\x17.tecton_proto.common.IdR\rfeatureViewId\x12*\n\x11\x66\x65\x61ture_view_name\x18\x02 \x01(\tR\x0f\x66\x65\x61tureViewName\x12\x39\n\ncreated_at\x18\x03 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\tcreatedAt\x12H\n\x12\x65xpected_freshness\x18\x04 \x01(\x0b\x32\x19.google.protobuf.DurationR\x11\x65xpectedFreshness\x12\x1b\n\tis_stream\x18\x05 \x01(\x08R\x08isStream\x12\x37\n\x17materialization_enabled\x18\x06 \x01(\x08R\x16materializationEnabled\x12\x37\n\tfreshness\x18\x07 \x01(\x0b\x32\x19.google.protobuf.DurationR\tfreshness\x12\x19\n\x08is_stale\x18\x08 \x01(\x08R\x07isStaleB\x13\n\x0f\x63om.tecton.dataP\x01')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tecton_proto.data.freshness_status_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\017com.tecton.dataP\001'
  _FRESHNESSSTATUS._serialized_start=159
  _FRESHNESSSTATUS._serialized_end=588
# @@protoc_insertion_point(module_scope)



import enum
import logging

from tecton._internals import errors
from tecton._internals.metadata_service_impl import auth_lib
from tecton._internals.metadata_service_impl import trace
from tecton_core.errors import FailedPreconditionError
from tecton_core.errors import TectonAPIInaccessibleError
from tecton_core.errors import TectonAPIValidationError
from tecton_core.errors import TectonNotFoundError
from tecton_core.logger import get_logging_level


class gRPCStatus(enum.Enum):
    """gRPC response status codes.

    Status codes are replicated here to avoid importing the `grpc.StatusCode` enum class,
    which requires the grpcio library.

    https://grpc.github.io/grpc/core/md_doc_statuscodes.html
    """

    OK = 0
    CANCELLED = 1
    UNKNOWN = 2
    INVALID_ARGUMENT = 3
    DEADLINE_EXCEEDED = 4
    NOT_FOUND = 5
    ALREADY_EXISTS = 6
    PERMISSION_DENIED = 7
    RESOURCE_EXHAUSTED = 8
    FAILED_PRECONDITION = 9
    ABORTED = 10
    OUT_OF_RANGE = 11
    UNIMPLEMENTED = 12
    INTERNAL = 13
    UNAVAILABLE = 14
    DATA_LOSS = 15
    UNAUTHENTICATED = 16


def raise_for_grpc_status(status_code: int, details: str, host_url: str):
    """
    Raise an exception based on a gRPC error status code.
    """

    if status_code == gRPCStatus.OK.value:
        return

    # Error handling
    if status_code == gRPCStatus.UNAVAILABLE.value:
        raise TectonAPIInaccessibleError(details, host_url)

    if status_code == gRPCStatus.INVALID_ARGUMENT.value:
        raise TectonAPIValidationError(details)

    if status_code == gRPCStatus.FAILED_PRECONDITION.value:
        raise FailedPreconditionError(details)

    if status_code == gRPCStatus.UNAUTHENTICATED.value:
        raise PermissionError(
            f"Tecton credentials are invalid, not configured, or expired ({details}). "
            "To authenticate using an API key, set TECTON_API_KEY in your environment or use tecton.set_credentials(tecton_api_key=<key>). "
            "To authenticate using the CLI, run `tecton login`."
        )

    if status_code == gRPCStatus.PERMISSION_DENIED.value:
        if not auth_lib.request_has_token():
            # Remove this case in https://tecton.atlassian.net/browse/TEC-9107
            raise PermissionError(
                "Tecton credentials have insufficient permissions. "
                "To authenticate using an API key, set TECTON_API_KEY in your environment or use tecton.set_credentials(tecton_api_key=<key>). "
                "To authenticate using the CLI, run `tecton login`."
            )
        elif details != None and "InvalidToken" in details:
            # Remove this case in https://tecton.atlassian.net/browse/TEC-9107
            raise PermissionError(f"Configured Tecton credentials are not valid ({details}).")
        else:
            raise PermissionError(f"Insufficient permissions ({details}).")
    if status_code == gRPCStatus.NOT_FOUND.value:
        raise TectonNotFoundError(details)

    if get_logging_level() < logging.INFO:
        raise Exception(f"Unknown MDS exception. code={status_code}, details={details}")

    raise errors.INTERNAL_ERROR_FROM_MDS(details, trace.get_trace_id())



import enum
from typing import Any
from typing import Dict
from typing import List
from typing import Optional

import pypika
from pypika import AliasedQuery
from pypika import analytics
from pypika import CustomFunction
from pypika import Field
from pypika import Interval
from pypika import Table
from pypika.dialects import MySQLQuery
from pypika.dialects import PostgreSQLQuery
from pypika.functions import Cast
from pypika.functions import DateAdd
from pypika.queries import Selectable
from pypika.terms import AnalyticFunction
from pypika.terms import Function
from pypika.terms import LiteralValue
from pypika.terms import Term
from pypika.utils import format_alias_sql

from tecton_core import conf
from tecton_core.data_types import DataType
from tecton_proto.data.feature_store_pb2 import FeatureStoreFormatVersion


class CaseSensitiveSnowflakeQuery(pypika.queries.Query):
    @classmethod
    def _builder(cls, **kwargs: Any) -> "CaseSensitiveSnowflakeQueryBuilder":
        return CaseSensitiveSnowflakeQueryBuilder(**kwargs)


class CaseSensitiveSnowflakeQueryBuilder(pypika.dialects.SnowflakeQueryBuilder):
    QUOTE_CHAR = '"'


class CustomQuery(pypika.queries.QueryBuilder):
    """
    Defines a custom-query class. It's needed for us to wrap some user-defined sql string from transformations in a QueryBuilder object.
    """

    def __init__(self, sql: str):
        super().__init__(self)
        self.sql = sql
        self.withs: List[Selectable] = []

    def with_(self, selectable: Selectable, name: str):
        """
        overrides QueryBuilder.with_
        """
        t = AliasedQuery(name, selectable)
        self.withs.append(t)
        return self

    def get_sql(self, with_alias: bool = False, subquery: bool = False, **kwargs):
        """
        overrides QueryBuilder.get_sql
        """
        sql = ""
        if self.withs:
            sql += "WITH " + ",".join(
                clause.name + " AS (" + clause.get_sql(subquery=False, with_alias=False, **kwargs) + ") "
                for clause in self.withs
            )
        sql += self.sql
        if with_alias:
            self.alias = "sq0"
            sql = f"({sql})"
            return format_alias_sql(sql, self.alias or self._table_name, **kwargs)
        return sql


class LastValue(analytics.LastValue):
    """
    Fixed version of pypika's LastValue to handle Snowflake and Athena wanting "ignore nulls"
     to be outside the parens of the window func, and Spark using a bool param.
    """

    def get_special_params_sql(self, **kwargs: Any) -> Optional[str]:
        if dialect() == Dialect.SPARK:
            # see: https://docs.databricks.com/sql/language-manual/functions/last_value.html
            # for sparkSQL syntax
            return f", {self._ignore_nulls}"
        # Snowflake does not support adding a function param to indicate "ignore nulls"
        # It looks like LAST_VALUE(...) IGNORE NULLS OVER(...)
        # see: https://docs.snowflake.com/en/sql-reference/functions/last_value.html
        # Nor does Athena - Athena docs don't make this clear though.
        else:
            return None

    def get_function_sql(self, **kwargs: Any) -> str:
        if dialect() == Dialect.SPARK:
            return super(LastValue, self).get_function_sql(**kwargs)
        else:
            function_sql = super(AnalyticFunction, self).get_function_sql(**kwargs)
            partition_sql = self.get_partition_sql(**kwargs)

            sql = function_sql
            if self._ignore_nulls:
                sql += " IGNORE NULLS"
            if self._include_over:
                sql += " OVER({partition_sql})".format(partition_sql=partition_sql)
            return sql


class Dialect(str, enum.Enum):
    SNOWFLAKE = "snowflake"
    ATHENA = "athena"
    SPARK = "spark"


def dialect() -> Dialect:
    # TODO(Daryl) - infer the sql dialect from the fv itself,
    # We just use the conf for now as athena transformations are still under development.
    d = conf.get_or_none("SQL_DIALECT")
    try:
        return Dialect(d)
    except:
        raise Exception(f"Unsupported sql dialect: set SQL_DIALECT to {[x.value for x in Dialect]}")


def Query():
    if dialect() == Dialect.SNOWFLAKE:
        return CaseSensitiveSnowflakeQuery
    elif dialect() == Dialect.SPARK:
        # Spark (similar to HiveSQL) uses backticks like MySQL
        return MySQLQuery
    else:
        # Athena (similar to PrestoSQL) uses doublequotes like Postgres
        return PostgreSQLQuery


# Spark, Athena, and Snowflake use different versions of structs.
class Struct(Function):
    def __init__(self, dialect: Dialect, *args, **kwargs):
        if dialect == Dialect.SPARK:
            f = "struct"
        elif dialect == Dialect.ATHENA:
            f = "row"
        elif dialect == Dialect.SNOWFLAKE:
            f = "array_construct"
        else:
            raise NotImplementedError
        super(Struct, self).__init__(f, *args, **kwargs)


def struct(field_names: List[str]) -> Term:
    fields = [Field(n) for n in field_names]
    return Struct(dialect(), *fields)


def default_case(field_name: str) -> str:
    # Snowflake defaults to uppercase
    if dialect() == Dialect.SNOWFLAKE:
        return field_name.upper()
    else:
        return field_name


def struct_extract(name: str, field_names: List[str], aliases: List[str], schema: Dict[str, DataType]) -> List[Field]:
    """
    Returns list of fields that pull out all `field_names` from the struct object, and alias them to `aliases`
    """
    if dialect() == Dialect.SPARK:
        assert len(field_names) == len(aliases)
        return [getattr(Table(name), (field)).as_(alias) for (field, alias) in zip(field_names, aliases)]
    elif dialect() == Dialect.ATHENA:
        return [getattr(Table(name), f"field{idx}").as_(aliases[idx]) for idx in range(len(aliases))]
    elif dialect() == Dialect.SNOWFLAKE:
        get = CustomFunction("get", ["array", "idx"])
        res = []
        for idx in range(len(aliases)):
            val = get(Field(name), idx)
            if field_names[idx] in schema:
                feature_type = schema[field_names[idx]].sql_type
                val = Cast(val, feature_type)
            res.append(val.as_(aliases[idx]))
        return res
    else:
        raise NotImplementedError


def to_timestamp(time_str: str) -> Term:
    if dialect() == Dialect.SNOWFLAKE or dialect() == Dialect.SPARK:
        c = CustomFunction("to_timestamp", ["time_str"])
    else:
        c = CustomFunction("from_iso8601_timestamp", ["time_str"])
    return c(time_str)


def date_add(interval: str, amount: int, time_field: Term) -> Term:
    if dialect() == Dialect.SNOWFLAKE:
        # snowflake uses dateadd rather than date_add; https://docs.snowflake.com/en/sql-reference/functions/dateadd.html
        c = CustomFunction("dateadd", ["interval", "amount", "time_field"])
        # LiteralValue will not put quotes around.
        # So we get dateadd(second, ...)
        interval = LiteralValue(interval)
    elif dialect() == Dialect.SPARK:
        if interval == "second":
            return time_field + Interval(seconds=amount)
        else:
            raise NotImplementedError
    else:
        c = DateAdd
        # For Athena, see: https://trino.io/docs/current/functions/datetime.html
        interval = f"'{interval}'"
    return c(interval, amount, time_field)


def to_unixtime(timestamp: Term) -> Term:
    if dialect() == Dialect.SNOWFLAKE:
        c = CustomFunction("date_part", ["epoch", "timestamp"])
        return c(Field("epoch_second"), timestamp)
    elif dialect() == Dialect.SPARK:
        c = CustomFunction("unix_timestamp", ["timestamp"])
        return c(timestamp)
    else:
        c = CustomFunction("to_unixtime", ["timestamp"])
        return c(timestamp)


def from_unixtime(unix_timestamp: Term) -> Term:
    if dialect() == Dialect.SNOWFLAKE:
        c = CustomFunction("to_timestamp", ["unix_timestamp"])
        return c(unix_timestamp)
    elif dialect() == Dialect.ATHENA:
        c = CustomFunction("from_unixtime", ["unix_timestamp"])
        return c(unix_timestamp)
    else:
        c = CustomFunction("from_unixtime", ["unix_timestamp"])
        return Cast(c(unix_timestamp), "timestamp")


def convert_epoch_term_in_seconds(timestamp: Term, time_stamp_feature_store_format_version: FeatureStoreFormatVersion):
    """
    Converts an epoch term to a timestamp column
    :param timestamp: epoch column [V0 : Seconds, V1 : Nanoseconds]
    :param time_stamp_feature_store_format_version: Feature Store Format Version
    :return: epoch in seconds
    """
    if time_stamp_feature_store_format_version == FeatureStoreFormatVersion.FEATURE_STORE_FORMAT_VERSION_DEFAULT:
        return timestamp
    return timestamp / 1e9


def convert_epoch_seconds_to_feature_store_format_version(
    timestamp: Term, feature_store_format_version: FeatureStoreFormatVersion
):
    if feature_store_format_version == FeatureStoreFormatVersion.FEATURE_STORE_FORMAT_VERSION_DEFAULT:
        return timestamp
    return timestamp * 1e9



import os
from importlib import resources

from tecton_core.logger import get_logger

logger = get_logger("udf_jar")


def get_udf_jar_path():
    # Environment variable override for non-Python package execution contexts.
    if os.environ.get("TECTON_UDFS_JAR"):
        return os.environ.get("TECTON_UDFS_JAR")

    with resources.path("tecton_spark.jars", "tecton-udfs-spark-3.jar") as p:
        return str(p)



from jinja2 import Environment
from jinja2 import PackageLoader
from jinja2 import StrictUndefined

from tecton_core.aggregation_utils import get_aggregation_function_name
from tecton_proto.common import column_type_pb2


def athena_function(value):
    fn = get_aggregation_function_name(value)
    if fn == "mean":
        return "avg"
    return fn


def load_template(name):
    env = Environment(
        loader=PackageLoader("tecton_athena"),
        autoescape=False,
        undefined=StrictUndefined,
    )
    env.globals["column_type_pb2"] = column_type_pb2
    env.filters["athena_function"] = athena_function
    return env.get_template(name)



from collections import defaultdict
from dataclasses import dataclass
from datetime import datetime
from datetime import timedelta
from typing import Dict
from typing import List
from typing import Optional
from typing import Set
from typing import Tuple
from typing import Union

import numpy
import pandas
import pendulum
import sqlparse

import tecton_spark.offline_store
from tecton_athena import odfv_helper
from tecton_athena.athena_session import get_session
from tecton_athena.data_catalog_helper import generate_sql_table_from_pandas_df
from tecton_athena.data_catalog_helper import PARTITION_TYPE_DATESTR
from tecton_athena.data_catalog_helper import PARTITION_TYPE_UNIX_EPOCH_NS
from tecton_athena.data_catalog_helper import register_feature_view_as_athena_table_if_necessary
from tecton_athena.pipeline_helper import pipeline_to_sql_string
from tecton_athena.templates_utils import load_template
from tecton_core import logger as logger_lib
from tecton_core import time_utils
from tecton_core.errors import START_TIME_NOT_BEFORE_END_TIME
from tecton_core.errors import TectonAthenaNotImplementedError
from tecton_core.errors import TectonValidationError
from tecton_core.feature_definition_wrapper import FeatureDefinitionWrapper as FeatureDefinition
from tecton_core.feature_set_config import FeatureSetConfig
from tecton_core.materialization_context import BoundMaterializationContext
from tecton_core.offline_store import window_size_seconds
from tecton_proto.common import aggregation_function_pb2 as afpb
from tecton_proto.data import feature_view_pb2


logger = logger_lib.get_logger("AthenaConnector")

SECONDS_TO_NANOSECONDS = 1000 * 1000 * 1000

HISTORICAL_FEATURES_TEMPLATE = load_template("historical_features.sql")
TIME_LIMIT_TEMPLATE = load_template("time_limit.sql")
MATERIALIZATION_TILE_TEMPLATE = load_template("materialization_tile.sql")
PARTIAL_AGGREGATION_TEMPLATE = load_template("run_partial_aggregation.sql")
FULL_AGGREGATION_TEMPLATE = load_template("run_full_aggregation.sql")

# TODO: Add non-trivial aggregation types: Variance, Standard deviations and Last-N aggregations.
AGGREGATION_PLANS = {
    afpb.AGGREGATION_FUNCTION_SUM: {("SUM", "SUM")},
    afpb.AGGREGATION_FUNCTION_MIN: {("MIN", "MIN")},
    afpb.AGGREGATION_FUNCTION_MAX: {("MAX", "MAX")},
    afpb.AGGREGATION_FUNCTION_COUNT: {("COUNT", "COUNT")},
    afpb.AGGREGATION_FUNCTION_MEAN: {("COUNT", "COUNT"), ("MEAN", "AVG")},
}


@dataclass
class _FeatureSetItemInput:
    """A simplified version of FeatureSetItem which is passed to the SQL template."""

    name: str
    timestamp_key: str
    join_keys: Dict[str, str]
    features: List[str]
    sql: str
    aggregation: Optional[feature_view_pb2.TrailingTimeWindowAggregation]
    ttl_seconds: Optional[int]


def _format_sql(sql_str: str) -> str:
    return sqlparse.format(sql_str, reindent=True)


def _get_feature_selection_time_bounds_from_spine_time_range(
    feature_definition: FeatureDefinition,
    spine_min_ts: pendulum.DateTime = None,
    spine_max_ts: pendulum.DateTime = None,
):
    feature_end_time = spine_max_ts
    feature_start_time = spine_min_ts

    if feature_definition.is_temporal:
        # We have to select at least as far back as the serving_ttl before the minimum spine time
        feature_start_time = spine_min_ts - feature_definition.serving_ttl
    elif feature_definition.is_temporal_aggregate:
        # For aggregates, we need to subtract the max aggregation time window AND an additional tile interval
        # We need to add an additional tile interval, because the spine's minimum timestamp may be inbetween two tile anchors
        feature_start_time = (
            spine_min_ts
            - pendulum.Duration(seconds=feature_definition.max_aggregation_window.ToSeconds())
            - feature_definition.get_tile_interval
        )

    return feature_start_time, feature_end_time


def _get_feature_selection_time_bounds(
    feature_definition: FeatureDefinition,
    start_time: Optional[datetime] = None,
    end_time: Optional[datetime] = None,
):
    if start_time is None or (
        feature_definition.feature_start_timestamp is not None
        and start_time < feature_definition.feature_start_timestamp
    ):
        logger.warning(
            "Defaulting start_time to FeatureView's feature_start_time because start_time was either not set or smaller than FeatureView start_time"
        )
        feature_start_time = feature_definition.feature_start_timestamp
        return feature_start_time, end_time
    else:
        feature_start_time = start_time
        return feature_start_time, end_time


def get_historical_features_sql(
    spine: Optional[Union[pandas.DataFrame]],
    feature_set_config: FeatureSetConfig,
    timestamp_key: Optional[str],
    include_feature_view_timestamp_columns: bool = False,
    start_time: Optional[datetime] = None,
    end_time: Optional[datetime] = None,
) -> Tuple[str, Optional[str]]:
    feature_set_items = feature_set_config.definitions_and_configs
    input_items = []
    if spine is None:
        # Only feature view is supported when the spine is not provided.
        # Feature service should always provide the spine.
        # SDK methods should never fail this check
        assert len(feature_set_items) == 1

    # Get a list of all the join keys in the spine.
    spine_keys = {}
    for item in feature_set_items:
        fd = item.feature_definition

        if not fd.is_on_demand:
            join_keys = {key: value for key, value in item.join_keys}
            spine_keys.update(join_keys)

    spine_info = upload_spine(spine=spine, expected_spine_keys=list(spine_keys), expected_timestamp_key=timestamp_key)

    for item in feature_set_items:
        fd = item.feature_definition

        if fd.is_on_demand:
            continue

        offline_store_type = fd.offline_store_config.WhichOneof("store_type")

        if offline_store_type not in ("parquet", "delta"):
            raise TectonValidationError(f"Offline store {offline_store_type} is not supported")

        # Change the feature view name if it's for internal udf use.
        if item.namespace.startswith("_udf_internal"):
            name = item.namespace.upper()
        else:
            name = fd.name

        join_keys = {key: value for key, value in item.join_keys}
        features = [
            col_name
            for col_name in fd.view_schema.column_names()
            if col_name not in (list(join_keys.keys()) + [fd.timestamp_key])
        ]
        if len(fd.online_serving_index.join_keys) != len(fd.join_keys):
            raise TectonAthenaNotImplementedError("Wildcard is not supported for Athena")

        if spine_info is not None:
            feature_start_time, feature_end_time = _get_feature_selection_time_bounds_from_spine_time_range(
                feature_definition=fd, spine_min_ts=spine_info.min_ts, spine_max_ts=spine_info.max_ts
            )
        else:
            feature_start_time, feature_end_time = _get_feature_selection_time_bounds(
                feature_definition=fd, start_time=start_time, end_time=end_time
            )

        if feature_start_time is not None and feature_end_time is not None and feature_start_time >= feature_end_time:
            raise START_TIME_NOT_BEFORE_END_TIME(feature_start_time, feature_end_time)

        sql_str = fetch_features_and_join_spine_sql(
            feature_definition=fd,
            feature_start_time=feature_start_time,
            feature_end_time=feature_end_time,
            spine=spine_info.spine_sql if spine_info is not None else None,
            spine_timestamp_key=timestamp_key,
        )
        input_items.append(
            _FeatureSetItemInput(
                name=name,
                timestamp_key=fd.timestamp_key,
                join_keys=join_keys,
                features=features,
                sql=sql_str,
                aggregation=(fd.trailing_time_window_aggregation if fd.is_temporal_aggregate else None),
                ttl_seconds=(int(fd.serving_ttl.total_seconds()) if (fd.is_temporal or fd.is_feature_table) else None),
            )
        )
    if spine_info is not None:
        sql_str = HISTORICAL_FEATURES_TEMPLATE.render(
            feature_set_items=input_items,
            spine_timestamp_key=timestamp_key,
            spine_sql=spine_info.spine_sql,
            include_feature_view_timestamp_columns=include_feature_view_timestamp_columns,
            spine_keys=list(spine_keys),
            spine_contains_non_join_keys=spine_info.contains_non_join_key_columns,
        )

    return (_format_sql(sql_str), spine_info and spine_info.table)


def _timestamp_to_partition(timestamp: pendulum.DateTime, partition_size: timedelta, partition_type: str):
    if timestamp is None:
        return None

    partition_size_seconds = window_size_seconds(partition_size)

    if partition_type == PARTITION_TYPE_DATESTR:
        unix_timestamp_seconds = int(timestamp.timestamp())
        dt = pendulum.from_timestamp(time_utils.align_epoch_downwards(unix_timestamp_seconds, partition_size_seconds))
        partition_format = tecton_spark.offline_store._timestamp_formats(partition_size).python_format
        return dt.strftime(partition_format)
    elif partition_type == PARTITION_TYPE_UNIX_EPOCH_NS:
        unix_timestamp_ns = int(timestamp.timestamp() * SECONDS_TO_NANOSECONDS)
        return time_utils.align_epoch_downwards(unix_timestamp_ns, partition_size_seconds * SECONDS_TO_NANOSECONDS)
    else:
        raise Exception("Unexpected partition_type")


def _ensure_time_is_pendulum(time):
    if time is None:
        return None

    if isinstance(time, pendulum.DateTime):
        return time

    if isinstance(time, (int, float)):
        return pendulum.from_timestamp(time)

    if isinstance(time, datetime):
        return pendulum.instance(time)

    raise Exception(f"Unexpected Time type {type(time)}")


# Todo: Refactor. Fairly ugly function
def _feature_view_select_all_sql(feature_definition: FeatureDefinition, start_time, end_time):
    athena_table = register_feature_view_as_athena_table_if_necessary(feature_definition, session=get_session())

    source = None
    timestamp_expression = None

    start_time = _ensure_time_is_pendulum(start_time)
    end_time = _ensure_time_is_pendulum(end_time)
    partition_lower_bound = _timestamp_to_partition(
        start_time, athena_table.partition_by_interval_timedelta, athena_table.partition_type
    )
    partition_upper_bound = _timestamp_to_partition(
        end_time, athena_table.partition_by_interval_timedelta, athena_table.partition_type
    )

    select_columns = []

    # Note: It's pretty bad that the timestamp field column is handled differently between aggregates and non-aggregates
    if feature_definition.is_temporal_aggregate:
        # Let's not select _anchor_time - leads to ambiguous queries
        if tecton_spark.offline_store.ANCHOR_TIME in athena_table.columns:
            # Note - this condition isn't always true: If the anchor time is just a partition column but not a table column. That's the case for Parquet based TWAGs
            del athena_table.columns[tecton_spark.offline_store.ANCHOR_TIME]

        select_columns = list(athena_table.columns.keys())
        timestamp_expression = f"from_unixtime({tecton_spark.offline_store.ANCHOR_TIME} / ({SECONDS_TO_NANOSECONDS}))"
        select_columns.append(f"{timestamp_expression} as {feature_definition.timestamp_key}")
    else:
        timestamp_expression = feature_definition.timestamp_key
        select_columns = list(athena_table.columns.keys())

    if athena_table.partition_type == PARTITION_TYPE_DATESTR:
        if partition_lower_bound:
            partition_lower_bound = f"'{partition_lower_bound}'"
        if partition_upper_bound:
            partition_upper_bound = f"'{partition_upper_bound}'"

    return TIME_LIMIT_TEMPLATE.render(
        select_columns=select_columns,
        start_time=start_time,
        end_time=end_time,
        source=athena_table.table,
        timestamp_key=timestamp_expression,
        partition_column=athena_table.partition_by,
        partition_lower_bound=partition_lower_bound,
        partition_upper_bound=partition_upper_bound,
    )


def fetch_features_and_join_spine_sql(
    feature_definition: FeatureDefinition,
    # start is inclusive and end is exclusive
    feature_start_time: Optional[datetime] = None,
    feature_end_time: Optional[datetime] = None,
    aggregation_level: str = "full",
    # If spine is provided, it will be used to join with the output results.
    # Currently only work with full aggregation.
    spine: Optional[str] = None,
    spine_timestamp_key: Optional[str] = None,
    spine_keys: Optional[List[str]] = None,
    mock_sql_inputs: Optional[Dict[str, str]] = None,
) -> str:
    if not feature_definition.writes_to_offline_store:
        raise Exception(
            f"FeatureView {feature_definition.name} does not have offline materialization enabled. Cannot proceed. Please enable offline materialization."
        )

    materialized_sql = _feature_view_select_all_sql(
        feature_definition, start_time=feature_start_time, end_time=feature_end_time
    )

    if feature_definition.is_temporal_aggregate:
        if aggregation_level == "full":
            aggregated_sql_str = FULL_AGGREGATION_TEMPLATE.render(
                source=materialized_sql,
                join_keys=feature_definition.join_keys,
                aggregation=feature_definition.trailing_time_window_aggregation,
                timestamp_key=feature_definition.timestamp_key,
                name=feature_definition.name,
                spine=spine,
                spine_timestamp_key=spine_timestamp_key,
                spine_keys=spine_keys,
                batch_schedule=int(feature_definition.batch_materialization_schedule.total_seconds()),
            )
            return _format_sql(aggregated_sql_str)
        else:
            raise ValueError(f"Unsupported aggregation level: {aggregation_level}")

    else:
        return _format_sql(materialized_sql)


@dataclass
class _SpineInfo:
    """Information about the spine"""

    table: str
    timestamp_key: str
    min_ts: pendulum.DateTime
    max_ts: pendulum.DateTime
    spine_keys: List[str]
    contains_non_join_key_columns: bool

    @property
    def spine_sql(self):
        return f"SELECT * FROM {self.table}"


def upload_spine(
    spine: Optional[Union[pandas.DataFrame]], expected_spine_keys: List[str], expected_timestamp_key: str
) -> Optional[_SpineInfo]:
    if spine is None:
        return None

    if isinstance(spine, pandas.DataFrame):
        spine_columns = set([c.lower() for c in list(spine.columns)])
        expected_spine_columns = set([c.lower() for c in expected_spine_keys])
        expected_spine_columns.add(expected_timestamp_key.lower())

        non_key_columns_in_spine = spine_columns - expected_spine_columns
        missing_columns_in_spine = expected_spine_columns - spine_columns

        if len(missing_columns_in_spine) > 0:
            raise Exception(f"Expected to find the following columns in the spine: {missing_columns_in_spine}")

        spine_timestamp_type = spine.dtypes[expected_timestamp_key].type
        if spine_timestamp_type != pandas.Timestamp and spine_timestamp_type != numpy.datetime64:
            raise Exception(f"Spine timestamp column must be of type Timestamp. It's of type {spine_timestamp_type}")

        spine_min_ts = pendulum.instance(spine[expected_timestamp_key].min())
        spine_max_ts = pendulum.instance(spine[expected_timestamp_key].max())

        spine_table_name = get_session().get_spine_temp_table_name()
        generate_sql_table_from_pandas_df(df=spine, session=get_session(), table_name=spine_table_name)
        return _SpineInfo(
            table=spine_table_name,
            timestamp_key=expected_timestamp_key,
            min_ts=spine_min_ts,
            max_ts=spine_max_ts,
            spine_keys=expected_spine_keys,
            contains_non_join_key_columns=len(non_key_columns_in_spine) > 0,
        )
    else:
        raise TectonAthenaNotImplementedError("Only pandas based spines are currently supported")


def _drop_internal_columns_from_ghf_result(data_df: pandas.DataFrame) -> pandas.DataFrame:
    internal_columns = [c for c in data_df.columns if c.startswith("_udf_internal")]
    return data_df.drop(columns=internal_columns)


def get_historical_features(
    spine: Optional[Union[pandas.DataFrame]],
    feature_set_config: FeatureSetConfig,
    timestamp_key: str,
    include_feature_view_timestamp_columns: bool = False,
    start_time: Optional[datetime] = None,
    end_time: Optional[datetime] = None,
) -> pandas.DataFrame:
    sql_str, spine_table = get_historical_features_sql(
        spine=spine,
        feature_set_config=feature_set_config,
        timestamp_key=timestamp_key,
        include_feature_view_timestamp_columns=include_feature_view_timestamp_columns,
        start_time=start_time,
        end_time=end_time,
    )

    try:
        logger.info("Getting historical features. This may take a few minutes...")

        materialized_feature_data_df = get_session().read_sql(sql_str)
        odfvs = odfv_helper.get_odfvs_from_feature_set_config(feature_set_config)
        internal_feature_data_df = odfv_helper.run_and_append_on_demand_features_to_historical_data(
            materialized_feature_data_df, odfvs
        )
        # ODFVs dependent FVs may have pulled in internal columns that need to be dropped
        final_feature_data_df = _drop_internal_columns_from_ghf_result(internal_feature_data_df)

        return final_feature_data_df
    finally:
        if spine_table:
            get_session().sql(f"DROP TABLE IF EXISTS {spine_table}")


def generate_run_batch_sql(
    fd: FeatureDefinition,
    # start is inclusive and end is exclusive
    feature_start_time: datetime,
    feature_end_time: datetime,
    aggregation_level: str,
    mock_sql_inputs: Optional[Dict[str, str]],
) -> str:
    materialization_context = BoundMaterializationContext._create_internal(
        pendulum.instance(feature_start_time), pendulum.instance(feature_end_time), fd.batch_materialization_schedule
    )

    pipeline_sql = pipeline_to_sql_string(
        pipeline=fd.pipeline,
        data_sources=fd.data_sources,
        transformations=fd.transformations,
        materialization_context=materialization_context,
        mock_sql_inputs=mock_sql_inputs,
    )
    materialized_sql = get_materialization_query(
        feature_definition=fd,
        source_sql=pipeline_sql,
        feature_start_time=feature_start_time,
        feature_end_time=feature_end_time,
    )

    if not fd.is_temporal_aggregate:
        return materialized_sql
    else:
        if aggregation_level == "disabled":
            sql_str = TIME_LIMIT_TEMPLATE.render(
                source=pipeline_sql,
                timestamp_key=fd.timestamp_key,
                start_time=feature_start_time,
                end_time=feature_end_time,
            )
        elif aggregation_level == "partial":
            # Rename the output columns, and add tile start/end time columns
            sql_str = PARTIAL_AGGREGATION_TEMPLATE.render(
                source=materialized_sql,
                join_keys=fd.join_keys,
                aggregations=_get_feature_view_aggregations(fd),
                slide_interval=fd.aggregate_slide_interval,
                slide_interval_string=fd.get_aggregate_slide_interval_string,
                timestamp_key=fd.timestamp_key,
            )
        elif aggregation_level == "full":
            sql_str = FULL_AGGREGATION_TEMPLATE.render(
                source=materialized_sql,
                join_keys=fd.join_keys,
                aggregation=fd.trailing_time_window_aggregation,
                timestamp_key=fd.timestamp_key,
                name=fd.name,
                batch_schedule=int(fd.batch_materialization_schedule.total_seconds()),
            )
        else:
            raise ValueError(f"Unsupported aggregation level: {aggregation_level}")

        return _format_sql(sql_str)


def get_materialization_query(
    fd: FeatureDefinition,
    sql_query: str,
    feature_start_time: datetime,
    feature_end_time: datetime,
):
    if fd.is_temporal_aggregate:
        sql_query = MATERIALIZATION_TILE_TEMPLATE.render(
            source=sql_query,
            join_keys=fd.join_keys,
            aggregations=_get_feature_view_aggregations(fd),
            slide_interval=fd.aggregate_slide_interval,
            timestamp_key=fd.timestamp_key,
        )
    return _format_sql(
        TIME_LIMIT_TEMPLATE.render(
            source=sql_query,
            timestamp_key=fd.timestamp_key,
            start_time=feature_start_time,
            end_time=feature_end_time,
        )
    )


def _get_feature_view_aggregations(fd: FeatureDefinition) -> Dict[str, Set[str]]:
    aggregations = defaultdict(set)
    for agg_feature in fd.fv_spec.aggregate_features:
        aggregate_function = AGGREGATION_PLANS[agg_feature.function]
        aggregations[agg_feature.input_feature_name].update(aggregate_function)

    # Need to order the functions for deterministic results.
    for key, value in aggregations.items():
        aggregations[key] = sorted(value)

    return aggregations



from typing import Optional

import pendulum
from pyspark.sql import DataFrame
from pyspark.sql import functions

from tecton_core.logger import get_logger


logger = get_logger("Query Builder")


def _get_time_limits_of_dataframe(df: DataFrame, time_key: str) -> Optional[pendulum.Period]:
    """The returned range is inclusive at the beginning & exclusive at the end: [start, end)."""
    # Fetch lower and upper time bound of the spine so that we can demand the individual feature definitions
    # to limit the amount of data they fetch from the raw data sources.
    # Returns None if df is empty.
    collected_df = df.select(
        functions.min(df[time_key]).alias("time_start"), functions.max(df[time_key]).alias("time_end")
    ).collect()
    time_start = collected_df[0]["time_start"]
    time_end = collected_df[0]["time_end"]
    if time_start is None or time_end is None:
        return None

    # Need to add 1 microsecond to the end time, since the range is exclusive at the end, and we need
    # to make sure to include the very last feature value (in terms of the event timestamp).
    return pendulum.instance(time_end).add(microseconds=1) - pendulum.instance(time_start)



from datetime import datetime
from typing import List
from typing import Optional
from typing import Union

import attrs
import pendulum
import pyspark
from pyspark.sql import DataFrame
from pyspark.sql import functions
from pyspark.sql import SparkSession
from pyspark.sql.types import LongType
from pyspark.sql.types import TimestampType

import tecton_core
from tecton_core import conf
from tecton_proto.data.feature_view_pb2 import MaterializationTimeRangePolicy
from tecton_spark.query.node import SparkExecNode


TECTON_FEATURE_TIMESTAMP_VALIDATOR = "_tecton_feature_timestamp_validator"
SKIP_FEATURE_TIMESTAMP_VALIDATION_ENV = "SKIP_FEATURE_TIMESTAMP_VALIDATION"
TIMESTAMP_VALIDATOR_UDF_REGISTERED = False


logger = tecton_core.logger.get_logger("QueryTree")


def _apply_or_check_feature_data_time_limits(
    spark: SparkSession,
    feature_df: DataFrame,
    time_range_policy: MaterializationTimeRangePolicy,
    timestamp_key: str,
    feature_data_time_limits: Optional[pendulum.Period],
) -> DataFrame:
    if time_range_policy == MaterializationTimeRangePolicy.MATERIALIZATION_TIME_RANGE_POLICY_FAIL_IF_OUT_OF_RANGE:
        return _validate_feature_timestamps(spark, feature_df, feature_data_time_limits, timestamp_key)
    elif time_range_policy == MaterializationTimeRangePolicy.MATERIALIZATION_TIME_RANGE_POLICY_FILTER_TO_RANGE:
        return _filter_to_feature_data_time_limits(feature_df, feature_data_time_limits, timestamp_key)
    else:
        raise ValueError(f"Unhandled time range policy: {time_range_policy}")


def _filter_to_feature_data_time_limits(
    feature_df: DataFrame,
    feature_data_time_limits: Optional[pendulum.Period],
    timestamp_key: Optional[str],
) -> DataFrame:
    if feature_data_time_limits:
        feature_df = feature_df.filter(
            (feature_df[timestamp_key] >= feature_data_time_limits.start)
            & (feature_df[timestamp_key] < feature_data_time_limits.end)
        )

    return feature_df


def _ensure_timestamp_validation_udf_registered(spark):
    """
    Register the Spark UDF that is contained in the JAR files and that is part of passed Spark session.
    If the UDF was already registered by the previous calls, do nothing. This is to avoid calling the JVM
    registration code repeatedly, which can be flaky due to Spark. We cannot use `SHOW USER FUNCTIONS` because
    there is a bug in the AWS Glue Catalog implementation that omits the catalog ID.

    Jars are included the following way into the Spark session:
     - For materialization jobs scheduled by Orchestrator, they are included in the Job submission API.
       In this case, we always use the default Spark session of the spun-up Spark cluster.
     - For interactive execution (or remote over db-connect / livy), we always construct Spark session
       manually and include appropriate JARs ourselves.
    """
    global TIMESTAMP_VALIDATOR_UDF_REGISTERED
    if not TIMESTAMP_VALIDATOR_UDF_REGISTERED:
        udf_generator = spark.sparkContext._jvm.com.tecton.udfs.spark3.RegisterFeatureTimestampValidator()
        udf_generator.register(TECTON_FEATURE_TIMESTAMP_VALIDATOR)
        TIMESTAMP_VALIDATOR_UDF_REGISTERED = True


def _validate_feature_timestamps(
    spark: SparkSession,
    feature_df: DataFrame,
    feature_data_time_limits: Optional[pendulum.Period],
    timestamp_key: Optional[str],
) -> DataFrame:
    if conf.get_or_none(SKIP_FEATURE_TIMESTAMP_VALIDATION_ENV) is True:
        logger.info(
            f"Note: skipping the feature timestamp validation step because `SKIP_FEATURE_TIMESTAMP_VALIDATION` is set to true."
        )
        return feature_df

    if feature_data_time_limits:
        _ensure_timestamp_validation_udf_registered(spark)

        start_time_expr = f"to_timestamp('{feature_data_time_limits.start}')"
        # Registered feature timestamp validation UDF checks that each timestamp is within *closed* time interval: [start_time, end_time].
        # So we subtract 1 microsecond here, before passing time limits to the UDF.
        end_time_expr = f"to_timestamp('{feature_data_time_limits.end - pendulum.duration(microseconds=1)}')"
        filter_expr = f"{TECTON_FEATURE_TIMESTAMP_VALIDATOR}({timestamp_key}, {start_time_expr}, {end_time_expr}, '{timestamp_key}')"

        # Force the output of the UDF to be filtered on, so the UDF cannot be optimized away.
        feature_df = feature_df.where(filter_expr)

    return feature_df


@attrs.frozen
class CustomFilterSparkNode(SparkExecNode):
    input_node: SparkExecNode
    filter_str: str

    def _to_dataframe(self, spark: pyspark.sql.SparkSession) -> pyspark.sql.DataFrame:
        input_df = self.input_node.to_dataframe(spark)
        return input_df.filter(self.filter_str)


@attrs.frozen
class FeatureTimeFilterSparkNode(SparkExecNode):
    input_node: SparkExecNode
    feature_data_time_limits: pendulum.Period
    policy: MaterializationTimeRangePolicy
    timestamp_field: str

    def _to_dataframe(self, spark: pyspark.sql.SparkSession) -> pyspark.sql.DataFrame:
        input_df = self.input_node.to_dataframe(spark)
        return _apply_or_check_feature_data_time_limits(
            spark, input_df, self.policy, self.timestamp_field, self.feature_data_time_limits
        )


@attrs.frozen
class EntityFilterSparkNode(SparkExecNode):
    feature_data: SparkExecNode
    entities: SparkExecNode
    entity_cols: List[str]

    def _to_dataframe(self, spark: pyspark.sql.SparkSession) -> pyspark.sql.DataFrame:
        feature_df = self.feature_data.to_dataframe(spark)
        entities_df = self.entities.to_dataframe(spark)
        return feature_df.join(entities_df, how="inner", on=self.entity_cols).select(feature_df.columns)


@attrs.frozen
class RespectFeatureStartTimeSparkNode(SparkExecNode):
    input_node: SparkExecNode
    retrieval_time_col: str
    feature_start_time: Union[pendulum.datetime, int]
    features: List[str]
    feature_store_format_version: int

    def _to_dataframe(self, spark: pyspark.sql.SparkSession) -> pyspark.sql.DataFrame:
        ret = self.input_node.to_dataframe(spark)

        ts_col_datatype = ret.schema[self.retrieval_time_col].dataType
        if isinstance(self.feature_start_time, int):
            if not isinstance(ts_col_datatype, LongType):
                raise RuntimeError(
                    f"Invalid feature_start_time column type, expected LongType but got: {ts_col_datatype}"
                )
        elif isinstance(self.feature_start_time, datetime):
            if not isinstance(ts_col_datatype, TimestampType):
                raise RuntimeError(
                    f"Invalid feature_start_time column type, expected TimestampType but got: {ts_col_datatype}"
                )
        else:
            raise RuntimeError(f"Invalid feature_start_time type: {type(self.feature_start_time)}")

        cond = functions.col(self.retrieval_time_col) >= functions.lit(self.feature_start_time)
        # select all non-feature cols, and null out any features outside of feature start time
        project_list = [col for col in ret.columns if col not in self.features]
        for c in self.features:
            newcol = functions.when(cond, functions.col(f"`{c}`")).otherwise(functions.lit(None)).alias(c)
            project_list.append(newcol)
        return ret.select(project_list)


@attrs.frozen
class RespectTTLSparkNode(SparkExecNode):
    input_node: SparkExecNode
    retrieval_time_col: str
    expiration_time_col: str
    features: List[str]

    def _to_dataframe(self, spark: pyspark.sql.SparkSession) -> pyspark.sql.DataFrame:
        ret = self.input_node.to_dataframe(spark)
        cond = functions.col(self.retrieval_time_col) < functions.col(self.expiration_time_col)
        # select all non-feature cols, and null out any features outside of ttl
        project_list = [col for col in ret.columns if col not in self.features]
        for c in self.features:
            newcol = functions.when(cond, functions.col(c)).otherwise(functions.lit(None)).alias(c)
            project_list.append(newcol)
        return ret.select(project_list)


@attrs.frozen
class StreamWatermarkSparkNode(SparkExecNode):
    input_node: SparkExecNode
    time_column: str
    stream_watermark: str

    def _to_dataframe(self, spark: pyspark.sql.SparkSession) -> pyspark.sql.DataFrame:
        ret = self.input_node.to_dataframe(spark)
        return ret.withWatermark(self.time_column, self.stream_watermark)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/feature_server/configuration/feature_server_configuration.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import duration_pb2 as google_dot_protobuf_dot_duration__pb2
from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2
from tecton_proto.args import pipeline_pb2 as tecton__proto_dot_args_dot_pipeline__pb2
from tecton_proto.auth import acl_pb2 as tecton__proto_dot_auth_dot_acl__pb2
from tecton_proto.common import aggregation_function_pb2 as tecton__proto_dot_common_dot_aggregation__function__pb2
from tecton_proto.common import id_pb2 as tecton__proto_dot_common_dot_id__pb2
from tecton_proto.data import feature_service_pb2 as tecton__proto_dot_data_dot_feature__service__pb2
from tecton_proto.data import feature_view_pb2 as tecton__proto_dot_data_dot_feature__view__pb2
from tecton_proto.data import tecton_api_key_pb2 as tecton__proto_dot_data_dot_tecton__api__key__pb2
from tecton_proto.data import transformation_pb2 as tecton__proto_dot_data_dot_transformation__pb2
from tecton_proto.common import data_type_pb2 as tecton__proto_dot_common_dot_data__type__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\nLtecton_proto/feature_server/configuration/feature_server_configuration.proto\x12)tecton_proto.feature_server.configuration\x1a\x1egoogle/protobuf/duration.proto\x1a\x1fgoogle/protobuf/timestamp.proto\x1a tecton_proto/args/pipeline.proto\x1a\x1btecton_proto/auth/acl.proto\x1a.tecton_proto/common/aggregation_function.proto\x1a\x1ctecton_proto/common/id.proto\x1a\'tecton_proto/data/feature_service.proto\x1a$tecton_proto/data/feature_view.proto\x1a&tecton_proto/data/tecton_api_key.proto\x1a&tecton_proto/data/transformation.proto\x1a#tecton_proto/common/data_type.proto\"\xeb\x01\n\x0c\x46\x65\x61turesPlan\x12[\n\x0c\x66\x65\x61ture_plan\x18\x01 \x01(\x0b\x32\x36.tecton_proto.feature_server.configuration.FeaturePlanH\x00R\x0b\x66\x65\x61turePlan\x12x\n\x17on_demand_features_plan\x18\x02 \x01(\x0b\x32?.tecton_proto.feature_server.configuration.OnDemandFeaturesPlanH\x00R\x14onDemandFeaturesPlanB\x04\n\x02\x66p\"\xec\x01\n\x06\x43olumn\x12:\n\tdata_type\x18\x05 \x01(\x0b\x32\x1d.tecton_proto.common.DataTypeR\x08\x64\x61taType\x12\x35\n\x17\x66\x65\x61ture_view_space_name\x18\x02 \x01(\tR\x14\x66\x65\x61tureViewSpaceName\x12;\n\x1a\x66\x65\x61ture_service_space_name\x18\x03 \x01(\tR\x17\x66\x65\x61tureServiceSpaceName\x12,\n\x12\x66\x65\x61ture_view_index\x18\x04 \x01(\x05R\x10\x66\x65\x61tureViewIndexJ\x04\x08\x01\x10\x02\"\xab\x0b\n\x0b\x46\x65\x61turePlan\x12V\n\routput_column\x18\x04 \x01(\x0b\x32\x31.tecton_proto.feature_server.configuration.ColumnR\x0coutputColumn\x12V\n\rinput_columns\x18\x01 \x03(\x0b\x32\x31.tecton_proto.feature_server.configuration.ColumnR\x0cinputColumns\x12[\n\x14\x61ggregation_function\x18\x03 \x01(\x0e\x32(.tecton_proto.common.AggregationFunctionR\x13\x61ggregationFunction\x12n\n\x1b\x61ggregation_function_params\x18\x06 \x01(\x0b\x32..tecton_proto.common.AggregationFunctionParamsR\x19\x61ggregationFunctionParams\x12H\n\x12\x61ggregation_window\x18\x05 \x01(\x0b\x32\x19.google.protobuf.DurationR\x11\x61ggregationWindow\x12N\n\tjoin_keys\x18\x07 \x03(\x0b\x32\x31.tecton_proto.feature_server.configuration.ColumnR\x08joinKeys\x12_\n\x12wildcard_join_keys\x18\x08 \x03(\x0b\x32\x31.tecton_proto.feature_server.configuration.ColumnR\x10wildcardJoinKeys\x12\x1d\n\ntable_name\x18\t \x01(\tR\ttableName\x12|\n\x19\x64\x61ta_table_timestamp_type\x18\x0f \x01(\x0e\x32\x41.tecton_proto.feature_server.configuration.DataTableTimestampTypeR\x16\x64\x61taTableTimestampType\x12\x82\x01\n\x1bstatus_table_timestamp_type\x18\x10 \x01(\x0e\x32\x43.tecton_proto.feature_server.configuration.StatusTableTimestampTypeR\x18statusTableTimestampType\x12#\n\rtimestamp_key\x18\x0c \x01(\tR\x0ctimestampKey\x12<\n\x0cslide_period\x18\r \x01(\x0b\x32\x19.google.protobuf.DurationR\x0bslidePeriod\x12:\n\x0bserving_ttl\x18\x0e \x01(\x0b\x32\x19.google.protobuf.DurationR\nservingTtl\x12\x30\n\x14refresh_status_table\x18\x11 \x01(\x08R\x12refreshStatusTable\x12*\n\x11\x66\x65\x61ture_view_name\x18\x15 \x01(\tR\x0f\x66\x65\x61tureViewName\x12&\n\x0f\x66\x65\x61ture_view_id\x18\x17 \x01(\tR\rfeatureViewId\x12?\n\x1c\x66\x65\x61ture_store_format_version\x18\x12 \x01(\x05R\x19\x66\x65\x61tureStoreFormatVersion\x12T\n\x13online_store_params\x18\x13 \x01(\x0b\x32$.tecton_proto.data.OnlineStoreParamsR\x11onlineStoreParams\x12.\n\x12\x64\x65letionTimeWindow\x18\x16 \x01(\x03R\x12\x64\x65letionTimeWindowJ\x04\x08\x02\x10\x03J\x04\x08\n\x10\x0bJ\x04\x08\x0b\x10\x0cJ\x04\x08\x14\x10\x15\"g\n\x11\x46\x65\x61tureVectorPlan\x12R\n\x08\x66\x65\x61tures\x18\x01 \x03(\x0b\x32\x36.tecton_proto.feature_server.configuration.FeaturePlanR\x08\x66\x65\x61tures\"\xc1\x05\n\x14OnDemandFeaturesPlan\x12l\n\x19\x61rgs_from_request_context\x18\x02 \x03(\x0b\x32\x31.tecton_proto.feature_server.configuration.ColumnR\x16\x61rgsFromRequestContext\x12K\n\x07outputs\x18\x03 \x03(\x0b\x32\x31.tecton_proto.feature_server.configuration.ColumnR\x07outputs\x12\x83\x01\n\x12\x66\x65\x61ture_set_inputs\x18\x05 \x03(\x0b\x32U.tecton_proto.feature_server.configuration.OnDemandFeaturesPlan.FeatureSetInputsEntryR\x10\x66\x65\x61tureSetInputs\x12\x37\n\x08pipeline\x18\x06 \x01(\x0b\x32\x1b.tecton_proto.args.PipelineR\x08pipeline\x12K\n\x0ftransformations\x18\x07 \x03(\x0b\x32!.tecton_proto.data.TransformationR\x0ftransformations\x12*\n\x11\x66\x65\x61ture_view_name\x18\x08 \x01(\tR\x0f\x66\x65\x61tureViewName\x12&\n\x0f\x66\x65\x61ture_view_id\x18\t \x01(\tR\rfeatureViewId\x1a\x81\x01\n\x15\x46\x65\x61tureSetInputsEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12R\n\x05value\x18\x02 \x01(\x0b\x32<.tecton_proto.feature_server.configuration.FeatureVectorPlanR\x05value:\x02\x38\x01J\x04\x08\x01\x10\x02J\x04\x08\x04\x10\x05\"\x81\x01\n\rLoggingConfig\x12\x1f\n\x0bsample_rate\x18\x01 \x01(\x02R\nsampleRate\x12.\n\x13log_effective_times\x18\x02 \x01(\x08R\x11logEffectiveTimes\x12\x1f\n\x0b\x61vro_schema\x18\x03 \x01(\tR\navroSchema\"\x8d\x05\n\x12\x46\x65\x61tureServicePlan\x12\x45\n\x12\x66\x65\x61ture_service_id\x18\x04 \x01(\x0b\x32\x17.tecton_proto.common.IdR\x10\x66\x65\x61tureServiceId\x12?\n\x0f\x66\x65\x61ture_view_id\x18\n \x01(\x0b\x32\x17.tecton_proto.common.IdR\rfeatureViewId\x12\x30\n\x14\x66\x65\x61ture_service_name\x18\x02 \x01(\tR\x12\x66\x65\x61tureServiceName\x12*\n\x11\x66\x65\x61ture_view_name\x18\t \x01(\tR\x0f\x66\x65\x61tureViewName\x12%\n\x0eworkspace_name\x18\x08 \x01(\tR\rworkspaceName\x12^\n\x0e\x66\x65\x61tures_plans\x18\r \x03(\x0b\x32\x37.tecton_proto.feature_server.configuration.FeaturesPlanR\rfeaturesPlans\x12N\n\x11join_key_template\x18\x05 \x01(\x0b\x32\".tecton_proto.data.JoinKeyTemplateR\x0fjoinKeyTemplate\x12\x41\n\x10\x66\x65\x61ture_view_ids\x18\x06 \x03(\x0b\x32\x17.tecton_proto.common.IdR\x0e\x66\x65\x61tureViewIds\x12_\n\x0elogging_config\x18\x0c \x01(\x0b\x32\x38.tecton_proto.feature_server.configuration.LoggingConfigR\rloggingConfigJ\x04\x08\x01\x10\x02J\x04\x08\x03\x10\x04J\x04\x08\x07\x10\x08J\x04\x08\x0b\x10\x0c\"\x80\x04\n\x11GlobalTableConfig\x12?\n\x0f\x66\x65\x61ture_view_id\x18\x04 \x01(\x0b\x32\x17.tecton_proto.common.IdR\rfeatureViewId\x12<\n\x0cslide_period\x18\x02 \x01(\x0b\x32\x19.google.protobuf.DurationR\x0bslidePeriod\x12\x12\n\x04size\x18\x03 \x01(\x05R\x04size\x12\x82\x01\n\x1bstatus_table_timestamp_type\x18\x06 \x01(\x0e\x32\x43.tecton_proto.feature_server.configuration.StatusTableTimestampTypeR\x18statusTableTimestampType\x12\x30\n\x14refresh_status_table\x18\x07 \x01(\x08R\x12refreshStatusTable\x12?\n\x1c\x66\x65\x61ture_store_format_version\x18\x08 \x01(\x05R\x19\x66\x65\x61tureStoreFormatVersion\x12T\n\x13online_store_params\x18\t \x01(\x0b\x32$.tecton_proto.data.OnlineStoreParamsR\x11onlineStoreParamsJ\x04\x08\x01\x10\x02J\x04\x08\x05\x10\x06\"\x87\x01\n\x12\x46\x65\x61tureServiceAcls\x12\x45\n\x12\x66\x65\x61ture_service_id\x18\x01 \x01(\x0b\x32\x17.tecton_proto.common.IdR\x10\x66\x65\x61tureServiceId\x12*\n\x04\x61\x63ls\x18\x02 \x03(\x0b\x32\x16.tecton_proto.auth.AclR\x04\x61\x63ls\"b\n\rWorkspaceAcls\x12%\n\x0eworkspace_name\x18\x01 \x01(\tR\rworkspaceName\x12*\n\x04\x61\x63ls\x18\x02 \x03(\x0b\x32\x16.tecton_proto.auth.AclR\x04\x61\x63ls\"\xf5\x06\n\x1a\x46\x65\x61tureServerConfiguration\x12?\n\rcomputed_time\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\x0c\x63omputedTime\x12h\n\x10\x66\x65\x61ture_services\x18\x01 \x03(\x0b\x32=.tecton_proto.feature_server.configuration.FeatureServicePlanR\x0f\x66\x65\x61tureServices\x12\xa0\x01\n\x1bglobal_table_config_by_name\x18\x03 \x03(\x0b\x32\x62.tecton_proto.feature_server.configuration.FeatureServerConfiguration.GlobalTableConfigByNameEntryR\x17globalTableConfigByName\x12O\n\x13\x61uthorized_api_keys\x18\x04 \x03(\x0b\x32\x1f.tecton_proto.data.TectonApiKeyR\x11\x61uthorizedApiKeys\x12o\n\x14\x66\x65\x61ture_service_acls\x18\x05 \x03(\x0b\x32=.tecton_proto.feature_server.configuration.FeatureServiceAclsR\x12\x66\x65\x61tureServiceAcls\x12_\n\x0eworkspace_acls\x18\x06 \x03(\x0b\x32\x38.tecton_proto.feature_server.configuration.WorkspaceAclsR\rworkspaceAcls\x12[\n\x17\x61ll_online_store_params\x18\x07 \x03(\x0b\x32$.tecton_proto.data.OnlineStoreParamsR\x14\x61llOnlineStoreParams\x1a\x88\x01\n\x1cGlobalTableConfigByNameEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12R\n\x05value\x18\x02 \x01(\x0b\x32<.tecton_proto.feature_server.configuration.GlobalTableConfigR\x05value:\x02\x38\x01*\x90\x01\n\x16\x44\x61taTableTimestampType\x12%\n!DATA_TABLE_TIMESTAMP_TYPE_UNKNOWN\x10\x00\x12&\n\"DATA_TABLE_TIMESTAMP_TYPE_SORT_KEY\x10\x01\x12\'\n#DATA_TABLE_TIMESTAMP_TYPE_ATTRIBUTE\x10\x02*\x98\x01\n\x18StatusTableTimestampType\x12\'\n#STATUS_TABLE_TIMESTAMP_TYPE_UNKNOWN\x10\x00\x12(\n$STATUS_TABLE_TIMESTAMP_TYPE_SORT_KEY\x10\x01\x12)\n%STATUS_TABLE_TIMESTAMP_TYPE_ATTRIBUTE\x10\x02\x42l\n(com.tecton.feature_service.configurationP\x01Z>github.com/tecton-ai/tecton_proto/feature_server/configuration')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tecton_proto.feature_server.configuration.feature_server_configuration_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n(com.tecton.feature_service.configurationP\001Z>github.com/tecton-ai/tecton_proto/feature_server/configuration'
  _ONDEMANDFEATURESPLAN_FEATURESETINPUTSENTRY._options = None
  _ONDEMANDFEATURESPLAN_FEATURESETINPUTSENTRY._serialized_options = b'8\001'
  _FEATURESERVERCONFIGURATION_GLOBALTABLECONFIGBYNAMEENTRY._options = None
  _FEATURESERVERCONFIGURATION_GLOBALTABLECONFIGBYNAMEENTRY._serialized_options = b'8\001'
  _DATATABLETIMESTAMPTYPE._serialized_start=5699
  _DATATABLETIMESTAMPTYPE._serialized_end=5843
  _STATUSTABLETIMESTAMPTYPE._serialized_start=5846
  _STATUSTABLETIMESTAMPTYPE._serialized_end=5998
  _FEATURESPLAN._serialized_start=526
  _FEATURESPLAN._serialized_end=761
  _COLUMN._serialized_start=764
  _COLUMN._serialized_end=1000
  _FEATUREPLAN._serialized_start=1003
  _FEATUREPLAN._serialized_end=2454
  _FEATUREVECTORPLAN._serialized_start=2456
  _FEATUREVECTORPLAN._serialized_end=2559
  _ONDEMANDFEATURESPLAN._serialized_start=2562
  _ONDEMANDFEATURESPLAN._serialized_end=3267
  _ONDEMANDFEATURESPLAN_FEATURESETINPUTSENTRY._serialized_start=3126
  _ONDEMANDFEATURESPLAN_FEATURESETINPUTSENTRY._serialized_end=3255
  _LOGGINGCONFIG._serialized_start=3270
  _LOGGINGCONFIG._serialized_end=3399
  _FEATURESERVICEPLAN._serialized_start=3402
  _FEATURESERVICEPLAN._serialized_end=4055
  _GLOBALTABLECONFIG._serialized_start=4058
  _GLOBALTABLECONFIG._serialized_end=4570
  _FEATURESERVICEACLS._serialized_start=4573
  _FEATURESERVICEACLS._serialized_end=4708
  _WORKSPACEACLS._serialized_start=4710
  _WORKSPACEACLS._serialized_end=4808
  _FEATURESERVERCONFIGURATION._serialized_start=4811
  _FEATURESERVERCONFIGURATION._serialized_end=5696
  _FEATURESERVERCONFIGURATION_GLOBALTABLECONFIGBYNAMEENTRY._serialized_start=5560
  _FEATURESERVERCONFIGURATION_GLOBALTABLECONFIGBYNAMEENTRY._serialized_end=5696
# @@protoc_insertion_point(module_scope)



import pytest

from tecton import tecton_context
from tecton_spark.udf_jar import get_udf_jar_path


@pytest.fixture(scope="session")
def tecton_pytest_spark_session():
    try:
        from pyspark.sql import SparkSession
    except ImportError:
        pytest.fail("Cannot create a SparkSession if `pyspark` is not installed.")

    active_session = SparkSession.getActiveSession()
    if active_session:
        pytest.fail(
            f"Cannot create SparkSession `tecton_pytest_spark_session` when there is already an active session: {active_session.sparkContext.appName}"
        )

    try:
        spark = (
            SparkSession.builder.appName("tecton_pytest_spark_session")
            .config("spark.jars", get_udf_jar_path())
            # This short-circuit's Spark's attempt to auto-detect a hostname for the master address, which can lead to
            # errors on hosts with "unusual" hostnames that Spark believes are invalid.
            .config("spark.driver.host", "localhost")
            .config("spark.ui.enabled", "false")
            .getOrCreate()
        )
    except Exception as e:
        # Unfortunately we can't do much better than parsing the error message since Spark raises `Exception` rather than a more specific type.
        if str(e) == "Java gateway process exited before sending its port number":
            pytest.fail(
                "Failed to start Java process for Spark, perhaps Java isn't installed or the 'JAVA_HOME' environment variable is not set?"
            )

        raise e

    try:
        tecton_context.set_tecton_spark_session(spark)
        yield spark
    finally:
        spark.stop()



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


def require_minimum_pandas_version():
    """ Raise ImportError if minimum version of Pandas is not installed
    """
    # TODO(HyukjinKwon): Relocate and deduplicate the version specification.
    minimum_pandas_version = "0.23.2"

    from distutils.version import LooseVersion
    try:
        import pandas
        have_pandas = True
    except ImportError as error:
        have_pandas = False
        raised_error = error
    if not have_pandas:
        raise ImportError("Pandas >= %s must be installed; however, "
                          "it was not found." % minimum_pandas_version) from raised_error
    if LooseVersion(pandas.__version__) < LooseVersion(minimum_pandas_version):
        raise ImportError("Pandas >= %s must be installed; however, "
                          "your version was %s." % (minimum_pandas_version, pandas.__version__))


def require_minimum_pyarrow_version():
    """ Raise ImportError if minimum version of pyarrow is not installed
    """
    # TODO(HyukjinKwon): Relocate and deduplicate the version specification.
    minimum_pyarrow_version = "1.0.0"

    from distutils.version import LooseVersion
    import os
    try:
        import pyarrow
        have_arrow = True
    except ImportError as error:
        have_arrow = False
        raised_error = error
    if not have_arrow:
        raise ImportError("PyArrow >= %s must be installed; however, "
                          "it was not found." % minimum_pyarrow_version) from raised_error
    if LooseVersion(pyarrow.__version__) < LooseVersion(minimum_pyarrow_version):
        raise ImportError("PyArrow >= %s must be installed; however, "
                          "your version was %s." % (minimum_pyarrow_version, pyarrow.__version__))
    if os.environ.get("ARROW_PRE_0_15_IPC_FORMAT", "0") == "1":
        raise RuntimeError("Arrow legacy IPC format is not supported in PySpark, "
                           "please unset ARROW_PRE_0_15_IPC_FORMAT")



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/data/feature_service.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from tecton_proto.validation import validator_pb2 as tecton__proto_dot_validation_dot_validator__pb2
from tecton_proto.common import data_type_pb2 as tecton__proto_dot_common_dot_data__type__pb2
from tecton_proto.args import feature_service_pb2 as tecton__proto_dot_args_dot_feature__service__pb2
from tecton_proto.data import fco_metadata_pb2 as tecton__proto_dot_data_dot_fco__metadata__pb2
from tecton_proto.common import id_pb2 as tecton__proto_dot_common_dot_id__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\'tecton_proto/data/feature_service.proto\x12\x11tecton_proto.data\x1a\'tecton_proto/validation/validator.proto\x1a#tecton_proto/common/data_type.proto\x1a\'tecton_proto/args/feature_service.proto\x1a$tecton_proto/data/fco_metadata.proto\x1a\x1ctecton_proto/common/id.proto\"\xe9\x03\n\x0e\x46\x65\x61tureService\x12\x45\n\x12\x66\x65\x61ture_service_id\x18\x01 \x01(\x0b\x32\x17.tecton_proto.common.IdR\x10\x66\x65\x61tureServiceId\x12M\n\x11\x66\x65\x61ture_set_items\x18\x02 \x03(\x0b\x32!.tecton_proto.data.FeatureSetItemR\x0f\x66\x65\x61tureSetItems\x12\x41\n\x0c\x66\x63o_metadata\x18\t \x01(\x0b\x32\x1e.tecton_proto.data.FcoMetadataR\x0b\x66\x63oMetadata\x12\x34\n\x16online_serving_enabled\x18\x0b \x01(\x08R\x14onlineServingEnabled\x12>\n\x07logging\x18\x0c \x01(\x0b\x32$.tecton_proto.args.LoggingConfigArgsR\x07logging\x12^\n\x0fvalidation_args\x18\r \x01(\x0b\x32\x35.tecton_proto.validation.FeatureServiceValidationArgsR\x0evalidationArgsJ\x04\x08\x03\x10\x04J\x04\x08\x04\x10\x05J\x04\x08\x05\x10\x06J\x04\x08\x06\x10\x07J\x04\x08\x07\x10\x08J\x04\x08\x08\x10\tJ\x04\x08\n\x10\x0b\"\xca\x01\n\x10JoinKeyComponent\x12*\n\x11spine_column_name\x18\x01 \x01(\tR\x0fspineColumnName\x12H\n\x0c\x62inding_type\x18\x02 \x01(\x0e\x32%.tecton_proto.data.JoinKeyBindingTypeR\x0b\x62indingType\x12:\n\tdata_type\x18\x04 \x01(\x0b\x32\x1d.tecton_proto.common.DataTypeR\x08\x64\x61taTypeJ\x04\x08\x03\x10\x04\"V\n\x0fJoinKeyTemplate\x12\x43\n\ncomponents\x18\x01 \x03(\x0b\x32#.tecton_proto.data.JoinKeyComponentR\ncomponents\"\x8a\x02\n\x0e\x46\x65\x61tureSetItem\x12?\n\x0f\x66\x65\x61ture_view_id\x18\x06 \x01(\x0b\x32\x17.tecton_proto.common.IdR\rfeatureViewId\x12\x62\n\x18join_configuration_items\x18\x03 \x03(\x0b\x32(.tecton_proto.data.JoinConfigurationItemR\x16joinConfigurationItems\x12\x1c\n\tnamespace\x18\x04 \x01(\tR\tnamespace\x12\'\n\x0f\x66\x65\x61ture_columns\x18\x05 \x03(\tR\x0e\x66\x65\x61tureColumnsJ\x04\x08\x01\x10\x02J\x06\x08\xe8\x07\x10\xe9\x07\"s\n\x15JoinConfigurationItem\x12*\n\x11spine_column_name\x18\x01 \x01(\tR\x0fspineColumnName\x12.\n\x13package_column_name\x18\x02 \x01(\tR\x11packageColumnName*|\n\x12JoinKeyBindingType\x12!\n\x1dJOIN_KEY_BINDING_TYPE_UNKNOWN\x10\x00\x12\x1f\n\x1bJOIN_KEY_BINDING_TYPE_BOUND\x10\x01\x12\"\n\x1eJOIN_KEY_BINDING_TYPE_WILDCARD\x10\x02\x42\x13\n\x0f\x63om.tecton.dataP\x01')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tecton_proto.data.feature_service_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\017com.tecton.dataP\001'
  _JOINKEYBINDINGTYPE._serialized_start=1420
  _JOINKEYBINDINGTYPE._serialized_end=1544
  _FEATURESERVICE._serialized_start=250
  _FEATURESERVICE._serialized_end=739
  _JOINKEYCOMPONENT._serialized_start=742
  _JOINKEYCOMPONENT._serialized_end=944
  _JOINKEYTEMPLATE._serialized_start=946
  _JOINKEYTEMPLATE._serialized_end=1032
  _FEATURESETITEM._serialized_start=1035
  _FEATURESETITEM._serialized_end=1301
  _JOINCONFIGURATIONITEM._serialized_start=1303
  _JOINCONFIGURATIONITEM._serialized_end=1418
# @@protoc_insertion_point(module_scope)



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from pyspark import since
from pyspark.ml.param import Params
from pyspark.ml.param.shared import HasCheckpointInterval, HasSeed, HasWeightCol, Param, \
    TypeConverters, HasMaxIter, HasStepSize, HasValidationIndicatorCol
from pyspark.ml.wrapper import JavaPredictionModel
from pyspark.ml.common import inherit_doc


@inherit_doc
class _DecisionTreeModel(JavaPredictionModel):
    """
    Abstraction for Decision Tree models.

    .. versionadded:: 1.5.0
    """

    @property
    @since("1.5.0")
    def numNodes(self):
        """Return number of nodes of the decision tree."""
        return self._call_java("numNodes")

    @property
    @since("1.5.0")
    def depth(self):
        """Return depth of the decision tree."""
        return self._call_java("depth")

    @property
    @since("2.0.0")
    def toDebugString(self):
        """Full description of model."""
        return self._call_java("toDebugString")

    @since("3.0.0")
    def predictLeaf(self, value):
        """
        Predict the indices of the leaves corresponding to the feature vector.
        """
        return self._call_java("predictLeaf", value)


class _DecisionTreeParams(HasCheckpointInterval, HasSeed, HasWeightCol):
    """
    Mixin for Decision Tree parameters.
    """

    leafCol = Param(Params._dummy(), "leafCol", "Leaf indices column name. Predicted leaf " +
                    "index of each instance in each tree by preorder.",
                    typeConverter=TypeConverters.toString)

    maxDepth = Param(Params._dummy(), "maxDepth", "Maximum depth of the tree. (>= 0) E.g., " +
                     "depth 0 means 1 leaf node; depth 1 means 1 internal node + 2 leaf nodes.",
                     typeConverter=TypeConverters.toInt)

    maxBins = Param(Params._dummy(), "maxBins", "Max number of bins for discretizing continuous " +
                    "features.  Must be >=2 and >= number of categories for any categorical " +
                    "feature.", typeConverter=TypeConverters.toInt)

    minInstancesPerNode = Param(Params._dummy(), "minInstancesPerNode", "Minimum number of " +
                                "instances each child must have after split. If a split causes " +
                                "the left or right child to have fewer than " +
                                "minInstancesPerNode, the split will be discarded as invalid. " +
                                "Should be >= 1.", typeConverter=TypeConverters.toInt)

    minWeightFractionPerNode = Param(Params._dummy(), "minWeightFractionPerNode", "Minimum "
                                     "fraction of the weighted sample count that each child "
                                     "must have after split. If a split causes the fraction "
                                     "of the total weight in the left or right child to be "
                                     "less than minWeightFractionPerNode, the split will be "
                                     "discarded as invalid. Should be in interval [0.0, 0.5).",
                                     typeConverter=TypeConverters.toFloat)

    minInfoGain = Param(Params._dummy(), "minInfoGain", "Minimum information gain for a split " +
                        "to be considered at a tree node.", typeConverter=TypeConverters.toFloat)

    maxMemoryInMB = Param(Params._dummy(), "maxMemoryInMB", "Maximum memory in MB allocated to " +
                          "histogram aggregation. If too small, then 1 node will be split per " +
                          "iteration, and its aggregates may exceed this size.",
                          typeConverter=TypeConverters.toInt)

    cacheNodeIds = Param(Params._dummy(), "cacheNodeIds", "If false, the algorithm will pass " +
                         "trees to executors to match instances with nodes. If true, the " +
                         "algorithm will cache node IDs for each instance. Caching can speed " +
                         "up training of deeper trees. Users can set how often should the cache " +
                         "be checkpointed or disable it by setting checkpointInterval.",
                         typeConverter=TypeConverters.toBoolean)

    def __init__(self):
        super(_DecisionTreeParams, self).__init__()

    def setLeafCol(self, value):
        """
        Sets the value of :py:attr:`leafCol`.
        """
        return self._set(leafCol=value)

    def getLeafCol(self):
        """
        Gets the value of leafCol or its default value.
        """
        return self.getOrDefault(self.leafCol)

    def getMaxDepth(self):
        """
        Gets the value of maxDepth or its default value.
        """
        return self.getOrDefault(self.maxDepth)

    def getMaxBins(self):
        """
        Gets the value of maxBins or its default value.
        """
        return self.getOrDefault(self.maxBins)

    def getMinInstancesPerNode(self):
        """
        Gets the value of minInstancesPerNode or its default value.
        """
        return self.getOrDefault(self.minInstancesPerNode)

    def getMinWeightFractionPerNode(self):
        """
        Gets the value of minWeightFractionPerNode or its default value.
        """
        return self.getOrDefault(self.minWeightFractionPerNode)

    def getMinInfoGain(self):
        """
        Gets the value of minInfoGain or its default value.
        """
        return self.getOrDefault(self.minInfoGain)

    def getMaxMemoryInMB(self):
        """
        Gets the value of maxMemoryInMB or its default value.
        """
        return self.getOrDefault(self.maxMemoryInMB)

    def getCacheNodeIds(self):
        """
        Gets the value of cacheNodeIds or its default value.
        """
        return self.getOrDefault(self.cacheNodeIds)


@inherit_doc
class _TreeEnsembleModel(JavaPredictionModel):
    """
    (private abstraction)
    Represents a tree ensemble model.
    """

    @property
    @since("2.0.0")
    def trees(self):
        """Trees in this ensemble. Warning: These have null parent Estimators."""
        return [_DecisionTreeModel(m) for m in list(self._call_java("trees"))]

    @property
    @since("2.0.0")
    def getNumTrees(self):
        """Number of trees in ensemble."""
        return self._call_java("getNumTrees")

    @property
    @since("1.5.0")
    def treeWeights(self):
        """Return the weights for each tree"""
        return list(self._call_java("javaTreeWeights"))

    @property
    @since("2.0.0")
    def totalNumNodes(self):
        """Total number of nodes, summed over all trees in the ensemble."""
        return self._call_java("totalNumNodes")

    @property
    @since("2.0.0")
    def toDebugString(self):
        """Full description of model."""
        return self._call_java("toDebugString")

    @since("3.0.0")
    def predictLeaf(self, value):
        """
        Predict the indices of the leaves corresponding to the feature vector.
        """
        return self._call_java("predictLeaf", value)


class _TreeEnsembleParams(_DecisionTreeParams):
    """
    Mixin for Decision Tree-based ensemble algorithms parameters.
    """

    subsamplingRate = Param(Params._dummy(), "subsamplingRate", "Fraction of the training data " +
                            "used for learning each decision tree, in range (0, 1].",
                            typeConverter=TypeConverters.toFloat)

    supportedFeatureSubsetStrategies = ["auto", "all", "onethird", "sqrt", "log2"]

    featureSubsetStrategy = \
        Param(Params._dummy(), "featureSubsetStrategy",
              "The number of features to consider for splits at each tree node. Supported " +
              "options: 'auto' (choose automatically for task: If numTrees == 1, set to " +
              "'all'. If numTrees > 1 (forest), set to 'sqrt' for classification and to " +
              "'onethird' for regression), 'all' (use all features), 'onethird' (use " +
              "1/3 of the features), 'sqrt' (use sqrt(number of features)), 'log2' (use " +
              "log2(number of features)), 'n' (when n is in the range (0, 1.0], use " +
              "n * number of features. When n is in the range (1, number of features), use" +
              " n features). default = 'auto'", typeConverter=TypeConverters.toString)

    def __init__(self):
        super(_TreeEnsembleParams, self).__init__()

    @since("1.4.0")
    def getSubsamplingRate(self):
        """
        Gets the value of subsamplingRate or its default value.
        """
        return self.getOrDefault(self.subsamplingRate)

    @since("1.4.0")
    def getFeatureSubsetStrategy(self):
        """
        Gets the value of featureSubsetStrategy or its default value.
        """
        return self.getOrDefault(self.featureSubsetStrategy)


class _RandomForestParams(_TreeEnsembleParams):
    """
    Private class to track supported random forest parameters.
    """

    numTrees = Param(Params._dummy(), "numTrees", "Number of trees to train (>= 1).",
                     typeConverter=TypeConverters.toInt)

    bootstrap = Param(Params._dummy(), "bootstrap", "Whether bootstrap samples are used "
                      "when building trees.", typeConverter=TypeConverters.toBoolean)

    def __init__(self):
        super(_RandomForestParams, self).__init__()

    @since("1.4.0")
    def getNumTrees(self):
        """
        Gets the value of numTrees or its default value.
        """
        return self.getOrDefault(self.numTrees)

    @since("3.0.0")
    def getBootstrap(self):
        """
        Gets the value of bootstrap or its default value.
        """
        return self.getOrDefault(self.bootstrap)


class _GBTParams(_TreeEnsembleParams, HasMaxIter, HasStepSize, HasValidationIndicatorCol):
    """
    Private class to track supported GBT params.
    """

    stepSize = Param(Params._dummy(), "stepSize",
                     "Step size (a.k.a. learning rate) in interval (0, 1] for shrinking " +
                     "the contribution of each estimator.",
                     typeConverter=TypeConverters.toFloat)

    validationTol = Param(Params._dummy(), "validationTol",
                          "Threshold for stopping early when fit with validation is used. " +
                          "If the error rate on the validation input changes by less than the " +
                          "validationTol, then learning will stop early (before `maxIter`). " +
                          "This parameter is ignored when fit without validation is used.",
                          typeConverter=TypeConverters.toFloat)

    @since("3.0.0")
    def getValidationTol(self):
        """
        Gets the value of validationTol or its default value.
        """
        return self.getOrDefault(self.validationTol)


class _HasVarianceImpurity(Params):
    """
    Private class to track supported impurity measures.
    """

    supportedImpurities = ["variance"]

    impurity = Param(Params._dummy(), "impurity",
                     "Criterion used for information gain calculation (case-insensitive). " +
                     "Supported options: " +
                     ", ".join(supportedImpurities), typeConverter=TypeConverters.toString)

    def __init__(self):
        super(_HasVarianceImpurity, self).__init__()

    @since("1.4.0")
    def getImpurity(self):
        """
        Gets the value of impurity or its default value.
        """
        return self.getOrDefault(self.impurity)


class _TreeClassifierParams(Params):
    """
    Private class to track supported impurity measures.

    .. versionadded:: 1.4.0
    """

    supportedImpurities = ["entropy", "gini"]

    impurity = Param(Params._dummy(), "impurity",
                     "Criterion used for information gain calculation (case-insensitive). " +
                     "Supported options: " +
                     ", ".join(supportedImpurities), typeConverter=TypeConverters.toString)

    def __init__(self):
        super(_TreeClassifierParams, self).__init__()

    @since("1.6.0")
    def getImpurity(self):
        """
        Gets the value of impurity or its default value.
        """
        return self.getOrDefault(self.impurity)


class _TreeRegressorParams(_HasVarianceImpurity):
    """
    Private class to track supported impurity measures.
    """
    pass



from abc import ABC
from abc import abstractmethod
from typing import List

from tecton_core import conf
from tecton_core.query.node_interface import NodeRef


class Rewrite(ABC):
    @abstractmethod
    def rewrite(self, node: NodeRef) -> NodeRef:
        raise NotImplementedError


# Mutates the input
def rewrite_tree(tree: NodeRef):
    if not conf.get_bool("QUERY_REWRITE_ENABLED"):
        return
    rewrites: List[Rewrite] = []
    for rewrite in rewrites:
        rewrite.rewrite(tree)



from typing import Optional
from typing import Tuple

import attrs
from typeguard import typechecked

from tecton_core.specs import tecton_object_spec
from tecton_core.specs import utils
from tecton_proto.args import entity_pb2 as entity__args_pb2
from tecton_proto.data import entity_pb2 as entity__data_pb2

__all__ = [
    "EntitySpec",
]


@utils.frozen_strict
class EntitySpec(tecton_object_spec.TectonObjectSpec):
    join_keys: Tuple[str, ...]

    # Temporarily expose the underlying data proto during migration.
    # TODO(TEC-12443): Remove this attribute.
    data_proto: Optional[entity__data_pb2.Entity] = attrs.field(metadata={utils.LOCAL_REMOTE_DIVERGENCE_ALLOWED: True})

    @classmethod
    @typechecked
    def from_data_proto(cls, proto: entity__data_pb2.Entity) -> "EntitySpec":
        return cls(
            metadata=tecton_object_spec.TectonObjectMetadataSpec.from_data_proto(proto.entity_id, proto.fco_metadata),
            join_keys=utils.get_tuple_from_repeated_field(proto.join_keys),
            data_proto=proto,
        )

    @classmethod
    @typechecked
    def from_args_proto(cls, proto: entity__args_pb2.EntityArgs) -> "EntitySpec":
        return cls(
            metadata=tecton_object_spec.TectonObjectMetadataSpec.from_args_proto(proto.entity_id, proto.info),
            join_keys=utils.get_tuple_from_repeated_field(proto.join_keys),
            data_proto=None,
        )
